import glob
import os
import shutil
import time
from pathlib import Path
from typing import List, Union

import numpy as np
import pandas as pd
import skimage.io
import yaml

"""
Global variable to save program call time.
"""

CALL_TIME = None


def read_parameters(parameter_file: str) -> dict:
    """Reads in default parameters and replaces user defined parameters.

    Args:
        parameter_file:
            Path to the parameter file.
    Returns:
        dictionary where all missing parameter values are replaced with the default values.

    """
    current_path = Path(os.path.dirname(os.path.realpath(__file__)))

    param_base_file = Path(current_path).joinpath("resources", "parameters.yml")

    with open(param_base_file, 'r') as yml_f:
        parameters = yaml.safe_load(yml_f)

    with open(parameter_file) as file:
        parameters_local = yaml.safe_load(file)

    # overwrite global parameters with local setting
    for key in parameters_local:
        parameters[key] = parameters_local[key]

    return parameters


def read_image(filename: Union[Path, str]) -> np.ndarray:
    """Reads an RGB or grayscale image with the scikit-learn library. Swaps axis if channels are not on last position.

    Args:
        filename:
            Path to the image file.

    Returns:
        image with channels on last position.
    """
    img_ = skimage.io.imread(filename)

    if len(img_.shape) <= 2:
        img_ = np.array([img_, img_])

    if img_.shape[0] < min(img_.shape[1], img_.shape[2]):
        print("Warning: channel is on the first dimension of the image.")
        img = np.swapaxes(np.swapaxes(img_, 0, 2), 0, 1)
    else:
        img = img_

    return img


def write_dict_to_yml(yml_file: Union[str, Path], d: dict) -> bool:
    """Writes a dictionary to a file in yml format.

    Args:
        yml_file:
            Path to the yml file.
        d:
            Dictionary to be written to the yml file.

    Returns:
        True if successful.

    """
    yml_file = Path(yml_file)
    p = Path(yml_file.parent)
    p.mkdir(parents=True, exist_ok=True)

    with open(yml_file, 'w+') as yml_f:
        yml_f.write(yaml.dump(d, Dumper=yaml.Dumper))

    return True


def create_path_recursively(path: Union[str, Path]) -> bool:
    """Creates a path. Creates missing parent folders.

    Args:
        path:
            Path to be created.

    Returns:
        True if successful.

    """
    p = Path(path)
    p.mkdir(parents=True, exist_ok=True)

    return True


def get_tif_list(path: Union[str, Path]) -> List[str]:
    """Get a list of all tif files in a folder.

    Args:
        path:
            Path to the folder.

    Returns:
        List of all tif files in the folder.

    """
    path = str(path)
    if not path.endswith(os.path.sep):
        path = path + os.path.sep

    return glob.glob(path + "*.tif")


def read_key_file(path: Union[str, Path]) -> pd.DataFrame:
    """Read the key file and return a pandas dataframe.

    Args:
        path:
            Path to the key file.

    Returns:
        Pandas dataframe with the key file.

    """
    return pd.read_csv(path)


def list_files_recursively(path: Union[str, Path], root: Union[str, Path] = None, relative: bool = False) -> List[Path]:
    """List all files in a folder recursively.

    Args:
        path:
            Path to the folder.
        root:
            Root path to be removed from the file path.
        relative:
            If True, the root path is removed from the file path.

    Returns:
        List of all files in the folder.

    """
    path_ = Path(path)
    if not root:
        root = path_
    files_list = []

    for cur_root, dirs, files in os.walk(path_):
        cur_root = Path(cur_root)

        for d in dirs:
            files_list += list_files_recursively(cur_root.joinpath(d), root, relative)
        for fi in files:
            if relative:
                files_list.append(cur_root.joinpath(fi).relative_to(root))
            else:
                files_list.append(cur_root.joinpath(fi))
        break

    return files_list


def get_doc_file_prefix() -> str:
    """Get the time when the program was called.

    Returns:
        Time when the program was called.

    """
    global CALL_TIME

    if not CALL_TIME:
        CALL_TIME = time.strftime('%Y%m%d_%H-%M-%S')

    call_time = CALL_TIME

    return "run_%s" % call_time


def copy(file: Union[str, Path], path_to: Union[str, Path]) -> Path:
    """Copies a file A to either folder B or file B. Makes sure folder structure for target exists.

    Args:
        file:
            Path to the file to be copied.
        path_to:
            Path to the target folder or file.

    Returns:
        Path to the copied file.

    """
    file = Path(file)
    path_to = Path(path_to)

    if os.path.exists(path_to) and os.path.samefile(file, path_to):
        return path_to

    create_path_recursively(path_to.parent)

    return Path(shutil.copy(file, path_to))



import copy
import math
import warnings
from os.path import basename as BASENAME

import numpy as np
import scipy.sparse
from scipy.sparse.csgraph import connected_components

# Copyright (c) 2007-2015, PySAL Developers
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# * Neither the name of the GeoDa Center for Geospatial Analysis and Computation
#   nor the names of its contributors may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

### copied form copied from pysal package
# see https://github.com/pysal/libpysal/blob/1d6326129904776b0f67597603f3d0af05a9da4c/libpysal/weights/weights.py
# pointing to a commit, not to main for reproducibility
# Removed classmethods from_file, from_shapefile, from_WSP, from_adjlist and method to_adjlist to
# avoid heavy dependencies.

class W(object):
    """
    Spatial weights class. Class attributes are described by their
    docstrings. to view, use the ``help`` function.

    Parameters
    ----------

    neighbors : dict
        Key is region ID, value is a list of neighbor IDS.
        For example, ``{'a':['b'],'b':['a','c'],'c':['b']}``.
    weights : dict
       Key is region ID, value is a list of edge weights.
       If not supplied all edge weights are assumed to have a weight of 1.
       For example, ``{'a':[0.5],'b':[0.5,1.5],'c':[1.5]}``.
    id_order : list
       An ordered list of ids, defines the order of observations when
       iterating over ``W`` if not set, lexicographical ordering is used
       to iterate and the ``id_order_set`` property will return ``False``.
       This can be set after creation by setting the ``id_order`` property.
    silence_warnings : bool
       By default ``libpysal`` will print a warning if the dataset contains
       any disconnected components or islands. To silence this warning set this
       parameter to ``True``.
    ids : list
        Values to use for keys of the neighbors and weights ``dict`` objects.

    Attributes
    ----------

    asymmetries
    cardinalities
    component_labels
    diagW2
    diagWtW
    diagWtW_WW
    histogram
    id2i
    id_order
    id_order_set
    islands
    max_neighbors
    mean_neighbors
    min_neighbors
    n
    n_components
    neighbor_offsets
    nonzero
    pct_nonzero
    s0
    s1
    s2
    s2array
    sd
    sparse
    trcW2
    trcWtW
    trcWtW_WW
    transform

    Examples
    --------

    >>> from libpysal.weights import W
    >>> neighbors = {0: [3, 1], 1: [0, 4, 2], 2: [1, 5], 3: [0, 6, 4], 4: [1, 3, 7, 5], 5: [2, 4, 8], 6: [3, 7], 7: [4, 6, 8], 8: [5, 7]}
    >>> weights = {0: [1, 1], 1: [1, 1, 1], 2: [1, 1], 3: [1, 1, 1], 4: [1, 1, 1, 1], 5: [1, 1, 1], 6: [1, 1], 7: [1, 1, 1], 8: [1, 1]}
    >>> w = W(neighbors, weights)
    >>> "%.3f"%w.pct_nonzero
    '29.630'

    Read from external `.gal file <https://geodacenter.github.io/workbook/4a_contig_weights/lab4a.html#gal-weights-file>`_.

    >>> import libpysal
    >>> w = libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
    >>> w.n
    78
    >>> "%.3f"%w.pct_nonzero
    '6.542'

    Set weights implicitly.

    >>> neighbors = {0: [3, 1], 1: [0, 4, 2], 2: [1, 5], 3: [0, 6, 4], 4: [1, 3, 7, 5], 5: [2, 4, 8], 6: [3, 7], 7: [4, 6, 8], 8: [5, 7]}
    >>> w = W(neighbors)
    >>> round(w.pct_nonzero,3)
    29.63
    >>> from libpysal.weights import lat2W
    >>> w = lat2W(100, 100)
    >>> w.trcW2
    39600.0
    >>> w.trcWtW
    39600.0
    >>> w.transform='r'
    >>> round(w.trcW2, 3)
    2530.722
    >>> round(w.trcWtW, 3)
    2533.667

    Cardinality Histogram:

    >>> w.histogram
    [(2, 4), (3, 392), (4, 9604)]

    Disconnected observations (islands):

    >>> from libpysal.weights import W
    >>> w = W({1:[0],0:[1],2:[], 3:[]})

    UserWarning: The weights matrix is not fully connected:
    There are 3 disconnected components.
    There are 2 islands with ids: 2, 3.

    """

    def __init__(
            self, neighbors, weights=None, id_order=None, silence_warnings=False, ids=None
    ):
        self.silence_warnings = silence_warnings
        self.transformations = {}
        self.neighbors = neighbors
        if not weights:
            weights = {}
            for key in neighbors:
                weights[key] = [1.0] * len(neighbors[key])
        self.weights = weights
        self.transformations["O"] = self.weights.copy()  # original weights
        self.transform = "O"
        if id_order is None:
            self._id_order = list(self.neighbors.keys())
            self._id_order.sort()
            self._id_order_set = False
        else:
            self._id_order = id_order
            self._id_order_set = True
        self._reset()
        self._n = len(self.weights)
        if (not self.silence_warnings) and (self.n_components > 1):
            message = (
                    "The weights matrix is not fully connected: "
                    "\n There are %d disconnected components." % self.n_components
            )
            ni = len(self.islands)
            if ni == 1:
                message = message + "\n There is 1 island with id: %s." % (
                    str(self.islands[0])
                )
            elif ni > 1:
                message = message + "\n There are %d islands with ids: %s." % (
                    ni,
                    ", ".join(str(island) for island in self.islands),
                )
            warnings.warn(message)

    def _reset(self):
        """Reset properties."""
        self._cache = {}

    def to_networkx(self):
        """Convert a weights object to a ``networkx`` graph.

        Returns
        -------
        A ``networkx`` graph representation of the ``W`` object.
        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError("NetworkX is required to use this function.")
        G = nx.DiGraph() if len(self.asymmetries) > 0 else nx.Graph()
        return nx.from_scipy_sparse_matrix(self.sparse, create_using=G)

    @classmethod
    def from_networkx(cls, graph, weight_col="weight"):
        """Convert a ``networkx`` graph to a PySAL ``W`` object.

        Parameters
        ----------
        graph : networkx.Graph
            The graph to convert to a ``W``.
        weight_col : string
            If the graph is labeled, this should be the name of the field
            to use as the weight for the ``W``.

        Returns
        -------
        w : libpysal.weights.W
            A ``W`` object containing the same graph as the ``networkx`` graph.
        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError("NetworkX is required to use this function.")
        sparse_matrix = nx.to_scipy_sparse_matrix(graph)
        w = WSP(sparse_matrix).to_W()
        return w

    @property
    def sparse(self):
        """Sparse matrix object. For any matrix manipulations required for w,
        ``w.sparse`` should be used. This is based on ``scipy.sparse``.
        """
        if "sparse" not in self._cache:
            self._sparse = self._build_sparse()
            self._cache["sparse"] = self._sparse
        return self._sparse

    @property
    def n_components(self):
        """Store whether the adjacency matrix is fully connected."""
        if "n_components" not in self._cache:
            self._n_components, self._component_labels = connected_components(
                self.sparse
            )
            self._cache["n_components"] = self._n_components
            self._cache["component_labels"] = self._component_labels
        return self._n_components

    @property
    def component_labels(self):
        """Store the graph component in which each observation falls."""
        if "component_labels" not in self._cache:
            self._n_components, self._component_labels = connected_components(
                self.sparse
            )
            self._cache["n_components"] = self._n_components
            self._cache["component_labels"] = self._component_labels
        return self._component_labels

    def _build_sparse(self):
        """Construct the sparse attribute."""

        row = []
        col = []
        data = []
        id2i = self.id2i
        for i, neigh_list in list(self.neighbor_offsets.items()):
            card = self.cardinalities[i]
            row.extend([id2i[i]] * card)
            col.extend(neigh_list)
            data.extend(self.weights[i])
        row = np.array(row)
        col = np.array(col)
        data = np.array(data)
        s = scipy.sparse.csr_matrix((data, (row, col)), shape=(self.n, self.n))
        return s

    @property
    def id2i(self):
        """Dictionary where the key is an ID and the value is that ID's
        index in ``W.id_order``.
        """
        if "id2i" not in self._cache:
            self._id2i = {}
            for i, id_i in enumerate(self._id_order):
                self._id2i[id_i] = i
            self._id2i = self._id2i
            self._cache["id2i"] = self._id2i
        return self._id2i

    @property
    def n(self):
        """Number of units."""
        if "n" not in self._cache:
            self._n = len(self.neighbors)
            self._cache["n"] = self._n
        return self._n

    @property
    def s0(self):
        r"""``s0`` is defined as

        .. math::

               s0=\sum_i \sum_j w_{i,j}

        """
        if "s0" not in self._cache:
            self._s0 = self.sparse.sum()
            self._cache["s0"] = self._s0
        return self._s0

    @property
    def s1(self):
        r"""``s1`` is defined as

        .. math::

               s1=1/2 \sum_i \sum_j \Big(w_{i,j} + w_{j,i}\Big)^2

        """
        if "s1" not in self._cache:
            t = self.sparse.transpose()
            t = t + self.sparse
            t2 = t.multiply(t)  # element-wise square
            self._s1 = t2.sum() / 2.0
            self._cache["s1"] = self._s1
        return self._s1

    @property
    def s2array(self):
        """Individual elements comprising ``s2``.

        See Also
        --------
        s2

        """
        if "s2array" not in self._cache:
            s = self.sparse
            self._s2array = np.array(s.sum(1) + s.sum(0).transpose()) ** 2
            self._cache["s2array"] = self._s2array
        return self._s2array

    @property
    def s2(self):
        r"""``s2`` is defined as

        .. math::

                s2=\sum_j \Big(\sum_i w_{i,j} + \sum_i w_{j,i}\Big)^2

        """
        if "s2" not in self._cache:
            self._s2 = self.s2array.sum()
            self._cache["s2"] = self._s2
        return self._s2

    @property
    def trcW2(self):
        """Trace of :math:`WW`.

        See Also
        --------
        diagW2

        """
        if "trcW2" not in self._cache:
            self._trcW2 = self.diagW2.sum()
            self._cache["trcw2"] = self._trcW2
        return self._trcW2

    @property
    def diagW2(self):
        """Diagonal of :math:`WW`.

        See Also
        --------
        trcW2

        """
        if "diagw2" not in self._cache:
            self._diagW2 = (self.sparse * self.sparse).diagonal()
            self._cache["diagW2"] = self._diagW2
        return self._diagW2

    @property
    def diagWtW(self):
        """Diagonal of :math:`W^{'}W`.

        See Also
        --------
        trcWtW

        """
        if "diagWtW" not in self._cache:
            self._diagWtW = (self.sparse.transpose() * self.sparse).diagonal()
            self._cache["diagWtW"] = self._diagWtW
        return self._diagWtW

    @property
    def trcWtW(self):
        """Trace of :math:`W^{'}W`.

        See Also
        --------
        diagWtW

        """
        if "trcWtW" not in self._cache:
            self._trcWtW = self.diagWtW.sum()
            self._cache["trcWtW"] = self._trcWtW
        return self._trcWtW

    @property
    def diagWtW_WW(self):
        """Diagonal of :math:`W^{'}W + WW`."""
        if "diagWtW_WW" not in self._cache:
            wt = self.sparse.transpose()
            w = self.sparse
            self._diagWtW_WW = (wt * w + w * w).diagonal()
            self._cache["diagWtW_WW"] = self._diagWtW_WW
        return self._diagWtW_WW

    @property
    def trcWtW_WW(self):
        """Trace of :math:`W^{'}W + WW`."""
        if "trcWtW_WW" not in self._cache:
            self._trcWtW_WW = self.diagWtW_WW.sum()
            self._cache["trcWtW_WW"] = self._trcWtW_WW
        return self._trcWtW_WW

    @property
    def pct_nonzero(self):
        """Percentage of nonzero weights."""
        if "pct_nonzero" not in self._cache:
            self._pct_nonzero = 100.0 * self.sparse.nnz / (1.0 * self._n ** 2)
            self._cache["pct_nonzero"] = self._pct_nonzero
        return self._pct_nonzero

    @property
    def cardinalities(self):
        """Number of neighbors for each observation."""
        if "cardinalities" not in self._cache:
            c = {}
            for i in self._id_order:
                c[i] = len(self.neighbors[i])
            self._cardinalities = c
            self._cache["cardinalities"] = self._cardinalities
        return self._cardinalities

    @property
    def max_neighbors(self):
        """Largest number of neighbors."""
        if "max_neighbors" not in self._cache:
            self._max_neighbors = max(self.cardinalities.values())
            self._cache["max_neighbors"] = self._max_neighbors
        return self._max_neighbors

    @property
    def mean_neighbors(self):
        """Average number of neighbors."""
        if "mean_neighbors" not in self._cache:
            self._mean_neighbors = np.mean(list(self.cardinalities.values()))
            self._cache["mean_neighbors"] = self._mean_neighbors
        return self._mean_neighbors

    @property
    def min_neighbors(self):
        """Minimum number of neighbors."""
        if "min_neighbors" not in self._cache:
            self._min_neighbors = min(self.cardinalities.values())
            self._cache["min_neighbors"] = self._min_neighbors
        return self._min_neighbors

    @property
    def nonzero(self):
        """Number of nonzero weights."""
        if "nonzero" not in self._cache:
            self._nonzero = self.sparse.nnz
            self._cache["nonzero"] = self._nonzero
        return self._nonzero

    @property
    def sd(self):
        """Standard deviation of number of neighbors."""
        if "sd" not in self._cache:
            self._sd = np.std(list(self.cardinalities.values()))
            self._cache["sd"] = self._sd
        return self._sd

    @property
    def asymmetries(self):
        """List of id pairs with asymmetric weights."""
        if "asymmetries" not in self._cache:
            self._asymmetries = self.asymmetry()
            self._cache["asymmetries"] = self._asymmetries
        return self._asymmetries

    @property
    def islands(self):
        """List of ids without any neighbors."""
        if "islands" not in self._cache:
            self._islands = [i for i, c in list(self.cardinalities.items()) if c == 0]
            self._cache["islands"] = self._islands
        return self._islands

    @property
    def histogram(self):
        """Cardinality histogram as a dictionary where key is the id and
        value is the number of neighbors for that unit.
        """
        if "histogram" not in self._cache:
            ct, bin = np.histogram(
                list(self.cardinalities.values()),
                list(range(self.min_neighbors, self.max_neighbors + 2)),
            )
            self._histogram = list(zip(bin, ct))
            self._cache["histogram"] = self._histogram
        return self._histogram

    def __getitem__(self, key):
        """Allow a dictionary like interaction with the weights class.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w = lat2W()

        >>> w[0] == dict({1: 1.0, 5: 1.0})
        True
        """
        return dict(list(zip(self.neighbors[key], self.weights[key])))

    def __iter__(self):
        """
        Support iteration over weights.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W(3,3)
        >>> for i,wi in enumerate(w):
        ...     print(i,wi[0])
        ...
        0 0
        1 1
        2 2
        3 3
        4 4
        5 5
        6 6
        7 7
        8 8
        >>>
        """
        for i in self._id_order:
            yield i, dict(list(zip(self.neighbors[i], self.weights[i])))

    def remap_ids(self, new_ids):
        """
        In place modification throughout ``W`` of id values from
        ``w.id_order`` to ``new_ids`` in all.

        Parameters
        ----------

        new_ids : list, numpy.ndarray
            Aligned list of new ids to be inserted. Note that first
            element of ``new_ids`` will replace first element of
            ``w.id_order``, second element of ``new_ids`` replaces second
            element of ``w.id_order`` and so on.

        Examples
        --------

        >>> from libpysal.weights import lat2W
        >>> w = lat2W(3, 3)
        >>> w.id_order
        [0, 1, 2, 3, 4, 5, 6, 7, 8]
        >>> w.neighbors[0]
        [3, 1]
        >>> new_ids = ['id%i'%id for id in w.id_order]
        >>> _ = w.remap_ids(new_ids)
        >>> w.id_order
        ['id0', 'id1', 'id2', 'id3', 'id4', 'id5', 'id6', 'id7', 'id8']
        >>> w.neighbors['id0']
        ['id3', 'id1']
        """

        old_ids = self._id_order
        if len(old_ids) != len(new_ids):
            raise Exception(
                "W.remap_ids: length of `old_ids` does not match             that of"
                " new_ids"
            )
        if len(set(new_ids)) != len(new_ids):
            raise Exception("W.remap_ids: list `new_ids` contains duplicates")
        else:
            new_neighbors = {}
            new_weights = {}
            old_transformations = self.transformations["O"].copy()
            new_transformations = {}
            for o, n in zip(old_ids, new_ids):
                o_neighbors = self.neighbors[o]
                o_weights = self.weights[o]
                n_neighbors = [new_ids[old_ids.index(j)] for j in o_neighbors]
                new_neighbors[n] = n_neighbors
                new_weights[n] = o_weights[:]
                new_transformations[n] = old_transformations[o]
            self.neighbors = new_neighbors
            self.weights = new_weights
            self.transformations["O"] = new_transformations

            id_order = [self._id_order.index(o) for o in old_ids]
            for i, id_ in enumerate(id_order):
                self.id_order[id_] = new_ids[i]

            self._reset()

    def __set_id_order(self, ordered_ids):
        """Set the iteration order in w. ``W`` can be iterated over. On
        construction the iteration order is set to the lexicographic order of
        the keys in the ``w.weights`` dictionary. If a specific order
        is required it can be set with this method.

        Parameters
        ----------

        ordered_ids : sequence
            Identifiers for observations in specified order.

        Notes
        -----

        The ``ordered_ids`` parameter is checked against the ids implied
        by the keys in ``w.weights``. If they are not equivalent sets an
        exception is raised and the iteration order is not changed.

        Examples
        --------

        >>> from libpysal.weights import lat2W
        >>> w=lat2W(3,3)
        >>> for i,wi in enumerate(w):
        ...     print(i, wi[0])
        ...
        0 0
        1 1
        2 2
        3 3
        4 4
        5 5
        6 6
        7 7
        8 8
        >>> w.id_order
        [0, 1, 2, 3, 4, 5, 6, 7, 8]
        >>> w.id_order=range(8,-1,-1)
        >>> list(w.id_order)
        [8, 7, 6, 5, 4, 3, 2, 1, 0]
        >>> for i,w_i in enumerate(w):
        ...     print(i,w_i[0])
        ...
        0 8
        1 7
        2 6
        3 5
        4 4
        5 3
        6 2
        7 1
        8 0

        """

        if set(self._id_order) == set(ordered_ids):
            self._id_order = ordered_ids
            self._id_order_set = True
            self._reset()
        else:
            raise Exception("ordered_ids do not align with W ids")

    def __get_id_order(self):
        """Returns the ids for the observations in the order in which they
        would be encountered if iterating over the weights.
        """
        return self._id_order

    id_order = property(__get_id_order, __set_id_order)

    @property
    def id_order_set(self):
        """Returns ``True`` if user has set ``id_order``, ``False`` if not.

        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W()
        >>> w.id_order_set
        True
        """
        return self._id_order_set

    @property
    def neighbor_offsets(self):
        """
        Given the current ``id_order``, ``neighbor_offsets[id]`` is the
        offsets of the id's neighbors in ``id_order``.

        Returns
        -------
        neighbor_list : list
            Offsets of the id's neighbors in ``id_order``.

        """

        if "neighbors_0" not in self._cache:
            self.__neighbors_0 = {}
            id2i = self.id2i
            for j, neigh_list in list(self.neighbors.items()):
                self.__neighbors_0[j] = [id2i[neigh] for neigh in neigh_list]
            self._cache["neighbors_0"] = self.__neighbors_0

        neighbor_list = self.__neighbors_0

        return neighbor_list

    def get_transform(self):
        """Getter for transform property.

        Returns
        -------
        transformation : str, None
            Valid transformation value. See the ``transform``
            parameters in ``set_transform()`` for a detailed description.

        See also
        --------
        set_transform

        """

        return self._transform

    def set_transform(self, value="B"):
        """Transformations of weights.

        Parameters
        ----------
        transform : str
            This parameter is not case sensitive. The following are
            valid transformations.

            * **B** -- Binary
            * **R** -- Row-standardization (global sum :math:`=n`)
            * **D** -- Double-standardization (global sum :math:`=1`)
            * **V** -- Variance stabilizing
            * **O** -- Restore original transformation (from instantiation)

        Notes
        -----

        Transformations are applied only to the value of the weights at
        instantiation. Chaining of transformations cannot be done on a ``W``
        instance.


        Examples
        --------
        >>> from libpysal.weights import lat2W
        >>> w=lat2W()
        >>> w.weights[0]
        [1.0, 1.0]
        >>> w.transform
        'O'
        >>> w.transform='r'
        >>> w.weights[0]
        [0.5, 0.5]
        >>> w.transform='b'
        >>> w.weights[0]
        [1.0, 1.0]
        """
        value = value.upper()
        self._transform = value
        if value in self.transformations:
            self.weights = self.transformations[value]
            self._reset()
        else:
            if value == "R":
                # row standardized weights
                weights = {}
                self.weights = self.transformations["O"]
                for i in self.weights:
                    wijs = self.weights[i]
                    row_sum = sum(wijs) * 1.0
                    if row_sum == 0.0:
                        if not self.silence_warnings:
                            print(("WARNING: ", i, " is an island (no neighbors)"))
                    weights[i] = [wij / row_sum for wij in wijs]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "D":
                # doubly-standardized weights
                # update current chars before doing global sum
                self._reset()
                s0 = self.s0
                ws = 1.0 / s0
                weights = {}
                self.weights = self.transformations["O"]
                for i in self.weights:
                    wijs = self.weights[i]
                    weights[i] = [wij * ws for wij in wijs]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "B":
                # binary transformation
                weights = {}
                self.weights = self.transformations["O"]
                for i in self.weights:
                    wijs = self.weights[i]
                    weights[i] = [1.0 for wij in wijs]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "V":
                # variance stabilizing
                weights = {}
                q = {}
                k = self.cardinalities
                s = {}
                Q = 0.0
                self.weights = self.transformations["O"]
                for i in self.weights:
                    wijs = self.weights[i]
                    q[i] = math.sqrt(sum([wij * wij for wij in wijs]))
                    s[i] = [wij / q[i] for wij in wijs]
                    Q += sum([si for si in s[i]])
                nQ = self.n / Q
                for i in self.weights:
                    weights[i] = [w * nQ for w in s[i]]
                weights = weights
                self.transformations[value] = weights
                self.weights = weights
                self._reset()
            elif value == "O":
                # put weights back to original transformation
                weights = {}
                original = self.transformations[value]
                self.weights = original
                self._reset()
            else:
                raise Exception("unsupported weights transformation")

    transform = property(get_transform, set_transform)

    def asymmetry(self, intrinsic=True):
        r"""
        Asymmetry check.

        Parameters
        ----------
        intrinsic : bool
            Default is ``True``. Intrinsic symmetry is defined as

            .. math::

                w_{i,j} == w_{j,i}

            If ``intrinsic`` is ``False`` symmetry is defined as

            .. math::

                i \in N_j \ \& \ j \in N_i

            where :math:`N_j` is the set of neighbors for :math:`j`.

        Returns
        -------
        asymmetries : list
            Empty if no asymmetries are found if asymmetries, then a
            ``list`` of ``(i,j)`` tuples is returned.

        Examples
        --------

        >>> from libpysal.weights import lat2W
        >>> w=lat2W(3,3)
        >>> w.asymmetry()
        []
        >>> w.transform='r'
        >>> w.asymmetry()
        [(0, 1), (0, 3), (1, 0), (1, 2), (1, 4), (2, 1), (2, 5), (3, 0), (3, 4), (3, 6), (4, 1), (4, 3), (4, 5), (4, 7), (5, 2), (5, 4), (5, 8), (6, 3), (6, 7), (7, 4), (7, 6), (7, 8), (8, 5), (8, 7)]
        >>> result = w.asymmetry(intrinsic=False)
        >>> result
        []
        >>> neighbors={0:[1,2,3], 1:[1,2,3], 2:[0,1], 3:[0,1]}
        >>> weights={0:[1,1,1], 1:[1,1,1], 2:[1,1], 3:[1,1]}
        >>> w=W(neighbors,weights)
        >>> w.asymmetry()
        [(0, 1), (1, 0)]
        """

        if intrinsic:
            wd = self.sparse.transpose() - self.sparse
        else:
            transform = self.transform
            self.transform = "b"
            wd = self.sparse.transpose() - self.sparse
            self.transform = transform

        ids = np.nonzero(wd)
        if len(ids[0]) == 0:
            return []
        else:
            ijs = list(zip(ids[0], ids[1]))
            ijs.sort()
            return ijs

    def symmetrize(self, inplace=False):
        """Construct a symmetric KNN weight. This ensures that the neighbors
        of each focal observation consider the focal observation itself as
        a neighbor. This returns a generic ``W`` object, since the object is no
        longer guaranteed to have ``k`` neighbors for each observation.
        """
        if not inplace:
            neighbors = copy.deepcopy(self.neighbors)
            weights = copy.deepcopy(self.weights)
            out_W = W(neighbors, weights, id_order=self.id_order)
            out_W.symmetrize(inplace=True)
            return out_W
        else:
            for focal, fneighbs in list(self.neighbors.items()):
                for j, neighbor in enumerate(fneighbs):
                    neighb_neighbors = self.neighbors[neighbor]
                    if focal not in neighb_neighbors:
                        self.neighbors[neighbor].append(focal)
                        self.weights[neighbor].append(self.weights[focal][j])
            self._cache = dict()
            return

    def full(self):
        """Generate a full ``numpy.ndarray``.

        Parameters
        ----------
        self : libpysal.weights.W
            spatial weights object

        Returns
        -------
        (fullw, keys) : tuple
            The first element being the full ``numpy.ndarray`` and second
            element keys being the ids associated with each row in the array.

        Examples
        --------
        >>> from libpysal.weights import W, full
        >>> neighbors = {'first':['second'],'second':['first','third'],'third':['second']}
        >>> weights = {'first':[1],'second':[1,1],'third':[1]}
        >>> w = W(neighbors, weights)
        >>> wf, ids = full(w)
        >>> wf
        array([[0., 1., 0.],
               [1., 0., 1.],
               [0., 1., 0.]])
        >>> ids
        ['first', 'second', 'third']
        """
        wfull = self.sparse.toarray()
        keys = list(self.neighbors.keys())
        if self.id_order:
            keys = self.id_order

        return (wfull, keys)

    def to_WSP(self):
        """Generate a ``WSP`` object.

        Returns
        -------

        implicit : libpysal.weights.WSP
            Thin ``W`` class

        See also
        --------
        WSP

        """
        return WSP(self.sparse, self._id_order)

    def set_shapefile(self, shapefile, idVariable=None, full=False):
        """
        Adding metadata for writing headers of ``.gal`` and ``.gwt`` files.

        Parameters
        ----------
        shapefile : str
            The shapefile name used to construct weights.
        idVariable : str
            The name of the attribute in the shapefile to associate
            with ids in the weights.
        full : bool
            Write out the entire path for a shapefile (``True``) or
            only the resources of the shapefile without extension (``False``).
            Default is ``True``.
        """

        if full:
            self._shpName = shapefile
        else:
            self._shpName = BASENAME(shapefile).split(".")[0]

        self._varName = idVariable

    def plot(
            self, gdf, indexed_on=None, ax=None, color="k", node_kws=None, edge_kws=None
    ):
        """Plot spatial weights objects. **Requires** ``matplotlib``, and
        implicitly requires a ``geopandas.GeoDataFrame`` as input.

        Parameters
        ----------
        gdf : geopandas.GeoDataFrame
            The original shapes whose topological relations are modelled in ``W``.
        indexed_on : str
            Column of ``geopandas.GeoDataFrame`` that the weights object uses
            as an index. Default is ``None``, so the index of the
            ``geopandas.GeoDataFrame`` is used.
        ax : matplotlib.axes.Axes
            Axis on which to plot the weights. Default is ``None``, so
            plots on the current figure.
        color : str
            ``matplotlib`` color string, will color both nodes and edges
            the same by default.
        node_kws : dict
            Keyword arguments dictionary to send to ``pyplot.scatter``,
            which provides fine-grained control over the aesthetics
            of the nodes in the plot.
        edge_kws : dict
            Keyword arguments dictionary to send to ``pyplot.plot``,
            which provides fine-grained control over the aesthetics
            of the edges in the plot.

        Returns
        -------
        f : matplotlib.figure.Figure
            Figure on which the plot is made.
        ax : matplotlib.axes.Axes
            Axis on which the plot is made.

        Notes
        -----
        If you'd like to overlay the actual shapes from the
        ``geopandas.GeoDataFrame``, call ``gdf.plot(ax=ax)`` after this.
        To plot underneath, adjust the z-order of the plot as follows:
        ``gdf.plot(ax=ax,zorder=0)``.

        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError(
                "W.plot depends on matplotlib.pyplot, and this was"
                "not able to be imported. \nInstall matplotlib to"
                "plot spatial weights."
            )
        if ax is None:
            f = plt.figure()
            ax = plt.gca()
        else:
            f = plt.gcf()
        if node_kws is not None:
            if "color" not in node_kws:
                node_kws["color"] = color
        else:
            node_kws = dict(color=color)
        if edge_kws is not None:
            if "color" not in edge_kws:
                edge_kws["color"] = color
        else:
            edge_kws = dict(color=color)

        for idx, neighbors in self:
            if idx in self.islands:
                continue
            if indexed_on is not None:
                neighbors = gdf[gdf[indexed_on].isin(neighbors)].index.tolist()
                idx = gdf[gdf[indexed_on] == idx].index.tolist()[0]
            centroids = gdf.loc[neighbors].centroid.apply(lambda p: (p.x, p.y))
            centroids = np.vstack(centroids.values)
            focal = np.hstack(gdf.loc[idx].geometry.centroid.xy)
            seen = set()
            for nidx, neighbor in zip(neighbors, centroids):
                if (idx, nidx) in seen:
                    continue
                ax.plot(*list(zip(focal, neighbor)), marker=None, **edge_kws)
                seen.update((idx, nidx))
                seen.update((nidx, idx))
        ax.scatter(
            gdf.centroid.apply(lambda p: p.x),
            gdf.centroid.apply(lambda p: p.y),
            **node_kws
        )
        return f, ax


class WSP(object):
    """Thin ``W`` class for ``spreg``.

    Parameters
    ----------

    sparse : scipy.sparse.{matrix-type}
        NxN object from ``scipy.sparse``

    Attributes
    ----------

    n           : int
                  description
    s0          : float
                  description
    trcWtW_WW   : float
                  description

    """

    def __init__(self, sparse, id_order=None, index=None):
        if not scipy.sparse.issparse(sparse):
            raise ValueError("must pass a scipy sparse object")
        rows, cols = sparse.shape
        if rows != cols:
            raise ValueError("Weights object must be square")
        self.sparse = sparse.tocsr()
        self.n = sparse.shape[0]
        self._cache = {}
        if id_order:
            if len(id_order) != self.n:
                raise ValueError(
                    "Number of values in id_order must match shape of sparse"
                )
            else:
                self._id_order = id_order
                self._cache["id_order"] = self._id_order
        # temp addition of index attribute
        import pandas as pd  # will be removed after refactoring is done

        if index is not None:
            if not isinstance(index, (pd.Index, pd.MultiIndex, pd.RangeIndex)):
                raise TypeError("index must be an instance of pandas.Index dtype")
            if len(index) != self.n:
                raise ValueError("Number of values in index must match shape of sparse")
        else:
            index = pd.RangeIndex(self.n)
        self.index = index

    @property
    def id_order(self):
        """An ordered list of ids, assumed to match the ordering in ``sparse``."""
        # Temporary solution until the refactoring is finished
        if "id_order" not in self._cache:
            if hasattr(self, "index"):
                self._id_order = self.index.tolist()
            else:
                self._id_order = list(range(self.n))
            self._cache["id_order"] = self._id_order
        return self._id_order

    @property
    def s0(self):
        r"""``s0`` is defined as:

        .. math::

               s0=\sum_i \sum_j w_{i,j}

        """
        if "s0" not in self._cache:
            self._s0 = self.sparse.sum()
            self._cache["s0"] = self._s0
        return self._s0

    @property
    def trcWtW_WW(self):
        """Trace of :math:`W^{'}W + WW`."""
        if "trcWtW_WW" not in self._cache:
            self._trcWtW_WW = self.diagWtW_WW.sum()
            self._cache["trcWtW_WW"] = self._trcWtW_WW
        return self._trcWtW_WW

    @property
    def diagWtW_WW(self):
        """Diagonal of :math:`W^{'}W + WW`."""
        if "diagWtW_WW" not in self._cache:
            wt = self.sparse.transpose()
            w = self.sparse
            self._diagWtW_WW = (wt * w + w * w).diagonal()
            self._cache["diagWtW_WW"] = self._diagWtW_WW
        return self._diagWtW_WW

    @classmethod
    def from_W(cls, W):
        """Constructs a ``WSP`` object from the ``W``'s sparse matrix.

        Parameters
        ----------
        W : libpysal.weights.W
            A PySAL weights object with a sparse form and ids.

        Returns
        -------
        A ``WSP`` instance.
        """
        return cls(W.sparse, id_order=W.id_order)

    def to_W(self, silence_warnings=False):
        """
        Convert a pysal WSP object (thin weights matrix) to a pysal W object.

        Parameters
        ----------
        self : WSP
            PySAL sparse weights object.
        silence_warnings : bool
            Switch to ``True`` to turn off print statements for every
            observation with islands. Default is ``False``, which does
            not silence warnings.

        Returns
        -------
        w : W
            PySAL weights object.

        """

        indices = list(self.sparse.indices)
        data = list(self.sparse.data)
        indptr = list(self.sparse.indptr)
        id_order = self.id_order
        if id_order:
            # replace indices with user IDs
            indices = [id_order[i] for i in indices]
        else:
            id_order = list(range(self.n))
        neighbors, weights = {}, {}
        start = indptr[0]
        for i in range(self.n):
            oid = id_order[i]
            end = indptr[i + 1]
            neighbors[oid] = indices[start:end]
            weights[oid] = data[start:end]
            start = end
        ids = copy.copy(self.id_order)
        w = W(neighbors, weights, ids, silence_warnings=silence_warnings)
        w._sparse = copy.deepcopy(self.sparse)
        w._cache["sparse"] = w._sparse
        return w



import glob
import platform
import sys
import unittest
from pathlib import Path

import pandas as pd

from polarityjam.argument_parsing import startup
from polarityjam.test.test_common import TestCommon
from polarityjam.utils.io import get_doc_file_prefix


class TestIntegration(TestCommon):

    def setUp(self) -> None:
        super().setUp()

    def check_doc_output(self, outputpath):
        outputpath = Path(outputpath)
        prefix = get_doc_file_prefix()

        log_file = outputpath.joinpath("%s%s" % (prefix, ".log"))
        param_file = outputpath.joinpath("%s%s" % (prefix, "_param.yml"))

        self.assertTrue(log_file.exists() and log_file.stat().st_size > 0, "Something is wrong with logging")
        self.assertTrue(
            param_file.exists() and param_file.stat().st_size > 0, "Something is wrong with copying the param file"
        )

    def test_run(self):
        # prepare
        in_file = str(self.get_test_image_path("060721_EGM2_18dyn_01.tif"))
        param_file = str(self.get_test_parameter_file("parameters_golgi_nuclei.yml"))
        out_path = str(self.output_path.joinpath("run"))

        # build arguments
        sys.argv = [sys.argv[0]] + ['run', param_file, in_file, out_path, '--filename_prefix=myfile']

        # call
        startup()

        # assert
        self.check_doc_output(out_path)

        # if version of cellpose is different, maybe labels change. Probably overall result not deterministic.
        # But staying in an interval is necessary. Hence, check number of segmented cells with error range of +-10%
        df = pd.read_csv(Path(out_path).joinpath("myfile.csv"))
        self.assertAlmostEqual(df.shape[0], 97, delta=10)

        # number of features should not change
        self.assertEqual(70, df.shape[1])

        # only one csv file in output
        num_csv = len(glob.glob(str(Path(out_path).joinpath("*.csv"))))
        self.assertEqual(1, num_csv)

    @unittest.skipIf(platform.system().lower() == 'windows', "Plotting too memory extensive. Skipping test!")
    def test_run_stack(self):
        in_path = str(self.get_test_image_folder("gn").joinpath("set_2"))
        param_file = str(self.get_test_parameter_file("parameters_golgi_nuclei.yml"))
        out_path = str(self.output_path.joinpath("run_stack"))

        # build arguments
        sys.argv = [sys.argv[0]] + ['run-stack', param_file, in_path, out_path]

        # call
        startup()

        # assert
        self.check_doc_output(out_path)

        df1 = pd.read_csv(Path(out_path).joinpath("060721_EGM2_18dyn_02.csv"))
        df2 = pd.read_csv(Path(out_path).joinpath("060721_EGM2_18dyn_03.csv"))

        self.assertAlmostEqual(107, df1.shape[0], delta=10)
        self.assertAlmostEqual(99, df2.shape[0], delta=10)

        # number of features should not change
        self.assertEqual(70, df1.shape[1])
        self.assertEqual(70, df2.shape[1])

        # two csv file in output
        num_csv = len(glob.glob(str(Path(out_path).joinpath("*.csv"))))
        self.assertEqual(3, num_csv)

    @unittest.skipIf(platform.system().lower() == 'windows', "Plotting too memory extensive. Skipping test!")
    def test_run_stack_no_golgi(self):
        in_path = str(self.get_test_image_folder("g"))
        param_file = str(self.get_test_parameter_file("parameters_no_golgi.yml"))
        out_path = str(self.output_path.joinpath("run_stack_no_golgi"))

        # build arguments
        sys.argv = [sys.argv[0]] + ['run-stack', param_file, in_path, out_path]

        # call
        startup()

        # assert
        self.check_doc_output(out_path)

        df1 = pd.read_csv(Path(out_path).joinpath("MAX_Flow_IF2_Vecad_Dapi_40X_I2.1.csv"))
        df2 = pd.read_csv(Path(out_path).joinpath("MAX_Flow_IF2_Vecad_Dapi_40X_I2.2.csv"))
        df3 = pd.read_csv(Path(out_path).joinpath("MAX_Flow_IF2_Vecad_Dapi_40X_I3.1.csv"))

        self.assertAlmostEqual(56, df1.shape[0], delta=5)
        self.assertAlmostEqual(56, df2.shape[0], delta=5)
        self.assertAlmostEqual(58, df3.shape[0], delta=7)

        # number of features should not change
        self.assertEqual(65, df1.shape[1])
        self.assertEqual(65, df2.shape[1])
        self.assertEqual(65, df3.shape[1])
        # three csv file in output
        num_csv = len(glob.glob(str(Path(out_path).joinpath("*.csv"))))
        self.assertEqual(4, num_csv)

    @unittest.skipIf(platform.system().lower() == 'windows', "Plotting too memory extensive. Skipping test!")
    def test_run_stack_no_nuclei(self):
        in_path = str(self.get_test_image_folder("n"))
        param_file = str(self.get_test_parameter_file("parameters_no_nuclei.yml"))
        out_path = str(self.output_path.joinpath("run_stack_no_nuclei"))

        # build arguments
        sys.argv = [sys.argv[0]] + ['run-stack', param_file, in_path, out_path]

        # call
        startup()

        # assert
        self.check_doc_output(out_path)

        # load csv
        df1 = pd.read_csv(Path(out_path).joinpath("MAX_8h_flow_uslide_new_setup_2021_10_14__21_19_47.csv"))
        df2 = pd.read_csv(Path(out_path).joinpath("MAX_8h_flow_uslide_new_setup_2021_10_14__21_29_39.csv"))
        df3 = pd.read_csv(Path(out_path).joinpath("MAX_8h_flow_uslide_new_setup_2021_10_14__21_36_01.csv"))

        # assert row count
        self.assertAlmostEqual(33, df1.shape[0], delta=5)
        self.assertAlmostEqual(28, df2.shape[0], delta=4)
        self.assertAlmostEqual(29, df3.shape[0], delta=4)

        # number of features should not change
        self.assertEqual(47, df1.shape[1])
        self.assertEqual(47, df2.shape[1])
        self.assertEqual(47, df3.shape[1])

        # three csv file in output
        num_csv = len(glob.glob(str(Path(out_path).joinpath("*.csv"))))
        self.assertEqual(4, num_csv)

    @unittest.skipIf(platform.system().lower() == 'windows', "Plotting too memory extensive. Skipping test!")
    def test_run_key(self):
        in_path = str(self.get_test_image_folder("gn"))
        in_key = str(self.get_test_key_file())
        param_file = str(self.get_test_parameter_file("parameters_golgi_nuclei.yml"))
        out_path = str(self.output_path.joinpath("run_key"))

        # build arguments
        sys.argv = [sys.argv[0]] + ['run-key', param_file, in_path, in_key, out_path]

        # call
        startup()

        # assert
        self.check_doc_output(out_path)

        # load summary csv
        df_sum = pd.read_csv(Path(out_path).joinpath("summary_table.csv"))

        # processed three images
        self.assertEqual(df_sum.shape, (3, 4))

        # processed two conditions
        Path(out_path).joinpath("cond_1", "merged_table_cond_1.csv").exists()
        Path(out_path).joinpath("cond_2", "merged_table_cond_2.csv").exists()



from polarityjam.polarityjam_logging import get_logger

PERMUTATIONS = 999
import numpy as np
import scipy.stats as stats

from polarityjam.model.weights import W


# ### copied from pysal package esda moran see https://pysal.org/esda/generated/esda.Moran.html
# or github https://github.com/pysal/esda/blob/c6a78c3fd4583ef7097970ff04a1d56b6dfbc7f7/esda/moran.py
# pointing to a commit, not to main for reproducibility
class Moran(object):
    """Moran's I Global Autocorrelation Statistic
    Parameters
    ----------
    y               : array
                      variable measured across n spatial units
    w               : W
                      spatial weights instance
    transformation  : string
                      weights transformation,  default is row-standardized "r".
                      Other options include "B": binary,  "D":
                      doubly-standardized,  "U": untransformed
                      (general weights), "V": variance-stabilizing.
    permutations    : int
                      number of random permutations for calculation of
                      pseudo-p_values
    two_tailed      : boolean
                      If True (default) analytical p-values for Moran are two
                      tailed, otherwise if False, they are one-tailed.
    Attributes
    ----------
    y            : array
                   original variable
    w            : W
                   original w object
    permutations : int
                   number of permutations
    I            : float
                   value of Moran's I
    EI           : float
                   expected value under normality assumption
    VI_norm      : float
                   variance of I under normality assumption
    seI_norm     : float
                   standard deviation of I under normality assumption
    z_norm       : float
                   z-value of I under normality assumption
    p_norm       : float
                   p-value of I under normality assumption
    VI_rand      : float
                   variance of I under randomization assumption
    seI_rand     : float
                   standard deviation of I under randomization assumption
    z_rand       : float
                   z-value of I under randomization assumption
    p_rand       : float
                   p-value of I under randomization assumption
    two_tailed   : boolean
                   If True p_norm and p_rand are two-tailed, otherwise they
                   are one-tailed.
    sim          : array
                   (if permutations>0)
                   vector of I values for permuted samples
    p_sim        : array
                   (if permutations>0)
                   p-value based on permutations (one-tailed)
                   null: spatial randomness
                   alternative: the observed I is extreme if
                   it is either extremely greater or extremely lower
                   than the values obtained based on permutations
    EI_sim       : float
                   (if permutations>0)
                   average value of I from permutations
    VI_sim       : float
                   (if permutations>0)
                   variance of I from permutations
    seI_sim      : float
                   (if permutations>0)
                   standard deviation of I under permutations.
    z_sim        : float
                   (if permutations>0)
                   standardized I based on permutations
    p_z_sim      : float
                   (if permutations>0)
                   p-value based on standard normal approximation from
                   permutations
    Notes
    -----
    Technical details and derivations can be found in :cite:`cliff81`.
    Examples
    --------
    >>> import libpysal
    >>> w = libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
    >>> f = libpysal.io.open(libpysal.examples.get_path("stl_hom.txt"))
    >>> y = np.array(f.by_col['HR8893'])
    >>> from esda.moran import Moran
    >>> mi = Moran(y,  w)
    >>> round(mi.I, 3)
    0.244
    >>> mi.EI
    -0.012987012987012988
    >>> mi.p_norm
    0.00027147862770937614
    SIDS example replicating OpenGeoda
    >>> w = libpysal.io.open(libpysal.examples.get_path("sids2.gal")).read()
    >>> f = libpysal.io.open(libpysal.examples.get_path("sids2.dbf"))
    >>> SIDR = np.array(f.by_col("SIDR74"))
    >>> mi = Moran(SIDR,  w)
    >>> round(mi.I, 3)
    0.248
    >>> mi.p_norm
    0.0001158330781489969
    One-tailed
    >>> mi_1 = Moran(SIDR,  w, two_tailed=False)
    >>> round(mi_1.I, 3)
    0.248
    >>> round(mi_1.p_norm, 4)
    0.0001
    """

    def __init__(
            self, y, w, transformation="r", permutations=PERMUTATIONS, two_tailed=True
    ):
        y = np.asarray(y).flatten()
        self.y = y
        w.transform = transformation
        self.w = w
        self.permutations = permutations
        self.__moments()
        self.I = self.__calc(self.z)
        self.z_norm = (self.I - self.EI) / self.seI_norm
        self.z_rand = (self.I - self.EI) / self.seI_rand

        if self.z_norm > 0:
            self.p_norm = 1 - stats.norm.cdf(self.z_norm)
            self.p_rand = 1 - stats.norm.cdf(self.z_rand)
        else:
            self.p_norm = stats.norm.cdf(self.z_norm)
            self.p_rand = stats.norm.cdf(self.z_rand)

        if two_tailed:
            self.p_norm *= 2.0
            self.p_rand *= 2.0

        if permutations:
            sim = [
                self.__calc(np.random.permutation(self.z)) for i in range(permutations)
            ]
            self.sim = sim = np.array(sim)
            above = sim >= self.I
            larger = above.sum()
            if (self.permutations - larger) < larger:
                larger = self.permutations - larger
            self.p_sim = (larger + 1.0) / (permutations + 1.0)
            self.EI_sim = sim.sum() / permutations
            self.seI_sim = np.array(sim).std()
            self.VI_sim = self.seI_sim ** 2
            if self.seI_sim != 0:
                self.z_sim = (self.I - self.EI_sim) / self.seI_sim
            else:
                self.z_sim = (self.I - self.EI_sim) / np.finfo(float).eps
            if self.z_sim > 0:
                self.p_z_sim = 1 - stats.norm.cdf(self.z_sim)
            else:
                self.p_z_sim = stats.norm.cdf(self.z_sim)

        # provide .z attribute that is znormalized
        sy = y.std()
        self.z /= sy

    def __moments(self):
        self.n = len(self.y)
        y = self.y
        z = y - y.mean()
        self.z = z
        self.z2ss = (z * z).sum()
        self.EI = -1.0 / (self.n - 1)
        n = self.n
        n2 = n * n
        s1 = self.w.s1
        s0 = self.w.s0
        s2 = self.w.s2
        s02 = s0 * s0
        v_num = n2 * s1 - n * s2 + 3 * s02
        v_den = (n - 1) * (n + 1) * s02
        self.VI_norm = v_num / v_den - (1.0 / (n - 1)) ** 2
        self.seI_norm = self.VI_norm ** (1 / 2.0)

        # variance under randomization
        xd4 = z ** 4
        xd2 = z ** 2
        k_num = xd4.sum() / n
        k_den = (xd2.sum() / n) ** 2
        k = k_num / k_den
        EI = self.EI
        A = n * ((n2 - 3 * n + 3) * s1 - n * s2 + 3 * s02)
        B = k * ((n2 - n) * s1 - 2 * n * s2 + 6 * s02)
        VIR = (A - B) / ((n - 1) * (n - 2) * (n - 3) * s02) - EI * EI
        self.VI_rand = VIR
        self.seI_rand = VIR ** (1 / 2.0)

    def __calc(self, z):
        zl = slag(self.w, z)
        inum = (z * zl).sum()
        return self.n / self.w.s0 * inum / self.z2ss

    @property
    def _statistic(self):
        """More consistent hidden attribute to access ESDA statistics"""
        return self.I

    @classmethod
    def by_col(
            cls, df, cols, w=None, inplace=False, pvalue="sim", outvals=None, **stat_kws
    ):
        """
        Function to compute a Moran statistic on a dataframe

        Arguments
        ---------
        df          :   pandas.DataFrame
                        a pandas dataframe with a geometry column
        cols        :   string or list of string
                        name or list of names of columns to use to compute the statistic
        w           :   pysal weights object
                        a weights object aligned with the dataframe. If not provided, this
                        is searched for in the dataframe's metadata
        inplace     :   bool
                        a boolean denoting whether to operate on the dataframe inplace or to
                        return a series contaning the results of the computation. If
                        operating inplace, the derived columns will be named
                        'column_moran'
        pvalue      :   string
                        a string denoting which pvalue should be returned. Refer to the
                        the Moran statistic's documentation for available p-values
        outvals     :   list of strings
                        list of arbitrary attributes to return as columns from the
                        Moran statistic
        **stat_kws  :   keyword arguments
                        options to pass to the underlying statistic. For this, see the
                        documentation for the Moran statistic.

        Returns
        --------
        If inplace, None, and operation is conducted on dataframe in memory. Otherwise,
        returns a copy of the dataframe with the relevant columns attached.

        """
        return _univariate_handler(
            df,
            cols,
            w=w,
            inplace=inplace,
            pvalue=pvalue,
            outvals=outvals,
            stat=cls,
            swapname=cls.__name__.lower(),
            **stat_kws
        )

# copied from pysal package
# see https://github.com/pysal/libpysal/blob/1d6326129904776b0f67597603f3d0af05a9da4c/libpysal/weights/spatial_lag.py
# pointing to a commit, not to main for reproducibility
# renamed from lag_spatial to slag
def slag(w, y):
    """
    Spatial lag operator.

    If w is row standardized, returns the average of each observation's neighbors;
    if not, returns the weighted sum of each observation's neighbors.

    Parameters
    ----------

    w                   : W
                          libpysal spatial weightsobject
    y                   : array
                          numpy array with dimensionality conforming to w (see examples)

    Returns
    -------

    wy                  : array
                          array of numeric values for the spatial lag

    Examples
    --------

    Setup a 9x9 binary spatial weights matrix and vector of data; compute the
    spatial lag of the vector.

    >>> import libpysal
    >>> import numpy as np
    >>> w = libpysal.weights.lat2W(3, 3)
    >>> y = np.arange(9)
    >>> yl = libpysal.weights.lag_spatial(w, y)
    >>> yl
    array([ 4.,  6.,  6., 10., 16., 14., 10., 18., 12.])

    Row standardize the weights matrix and recompute the spatial lag

    >>> w.transform = 'r'
    >>> yl = libpysal.weights.lag_spatial(w, y)
    >>> yl
    array([2.        , 2.        , 3.        , 3.33333333, 4.        ,
           4.66666667, 5.        , 6.        , 6.        ])


    Explicitly define data vector as 9x1 and recompute the spatial lag

    >>> y.shape = (9, 1)
    >>> yl = libpysal.weights.lag_spatial(w, y)
    >>> yl
    array([[2.        ],
           [2.        ],
           [3.        ],
           [3.33333333],
           [4.        ],
           [4.66666667],
           [5.        ],
           [6.        ],
           [6.        ]])


    Take the spatial lag of a 9x2 data matrix

    >>> yr = np.arange(8, -1, -1)
    >>> yr.shape = (9, 1)
    >>> x = np.hstack((y, yr))
    >>> yl = libpysal.weights.lag_spatial(w, x)
    >>> yl
    array([[2.        , 6.        ],
           [2.        , 6.        ],
           [3.        , 5.        ],
           [3.33333333, 4.66666667],
           [4.        , 4.        ],
           [4.66666667, 3.33333333],
           [5.        , 3.        ],
           [6.        , 2.        ],
           [6.        , 2.        ]])

    """
    return w.sparse * y


# copied from pysal package
# see https://github.com/pysal/esda/blob/c6a78c3fd4583ef7097970ff04a1d56b6dfbc7f7/esda/tabular.py
# pointing to a commit, not to main for reproducibility
def _univariate_handler(df, cols, stat=None, w=None, inplace=True,
                        pvalue='sim', outvals=None, swapname='', **kwargs):
    """
    Compute a univariate descriptive statistic `stat` over columns `cols` in
    `df`.

    Parameters
    ----------
    df          : pandas.DataFrame
                  the dataframe containing columns to compute the descriptive
                  statistics
    cols        : string or list of strings
                  one or more names of columns in `df` to use to compute
                  exploratory descriptive statistics.
    stat        : callable
                  a function that takes data as a first argument and any number
                  of configuration keyword arguments and returns an object
                  encapsulating the exploratory statistic results
    w           : pysal.weights.W
                  the spatial weights object corresponding to the dataframe
    inplace     : bool
                  a flag denoting whether to add the statistic to the dataframe
                  in memory, or to construct a copy of the dataframe and append
                  the results to the copy
    pvalue      : string
                  the name of the pvalue on the results object wanted
    outvals     : list of strings
                  names of attributes of the dataframe to attempt to flatten
                  into a column
    swapname    : string
                  suffix to replace generic identifier with. Each caller of this
                  function should set this to a unique column suffix
    **kwargs    : optional keyword arguments
                  options that are passed directly to the statistic
    """
    ### Preprocess
    if not inplace:
        new_df = df.copy()
        _univariate_handler(new_df, cols, stat=stat, w=w, pvalue=pvalue,
                            inplace=True, outvals=outvals,
                            swapname=swapname, **kwargs)
        return new_df
    if w is None:
        for name in df._metadata:
            this_obj = df.__dict__.get(name)
            if isinstance(this_obj, W):
                w = this_obj
    if w is None:
        raise Exception('Weights not provided and no weights attached to frame!'
                        ' Please provide a weight or attach a weight to the'
                        ' dataframe')
    ### Prep indexes
    if outvals is None:
        outvals = []
    outvals.insert(0, '_statistic')
    if pvalue.lower() in ['all', 'both', '*']:
        raise NotImplementedError("If you want more than one type of PValue,add"
                                  " the targeted pvalue type to outvals. For example:"
                                  " Geary(df, cols=['HOVAL'], w=w, outvals=['p_z_sim', "
                                  "'p_rand']")
    # this is nontrivial, since we
    # can't know which p_value types are on the object without computing it.
    # This is because we don't flag them with @properties, so they're just
    # arbitrarily assigned post-facto. One solution might be to post-process the
    # objects, determine which pvalue types are available, and then grab them
    # all if needed.

    if pvalue != '':
        outvals.append('p_' + pvalue.lower())
    if isinstance(cols, str):
        cols = [cols]

    ### Make closure around weights & apply columnwise
    def column_stat(column):
        return stat(column.values, w=w, **kwargs)

    stat_objs = df[cols].apply(column_stat)

    ### Assign into dataframe
    for col in cols:
        stat_obj = stat_objs[col]
        y = kwargs.get('y')
        if y is not None:
            col += '-' + y.name
        outcols = ['_'.join((col, val)) for val in outvals]
        for colname, attname in zip(outcols, outvals):
            df[colname] = stat_obj.__getattribute__(attname)
    if swapname != '':
        df.columns = [_swap_ending(col, swapname) if col.endswith('_statistic') else col
                      for col in df.columns]


# copied from pysal package
# see https://github.com/pysal/esda/blob/c6a78c3fd4583ef7097970ff04a1d56b6dfbc7f7/esda/tabular.py
# pointing to a commit, not to main for reproducibility
def _swap_ending(s, ending, delim='_'):
    """
    Replace the ending of a string, delimited into an arbitrary
    number of chunks by `delim`, with the ending provided

    Parameters
    ----------
    s       :   string
                string to replace endings
    ending  :   string
                string used to replace ending of `s`
    delim   :   string
                string that splits s into one or more parts

    Returns
    -------
    new string where the final chunk of `s`, delimited by `delim`, is replaced
    with `ending`.
    """
    parts = [x for x in s.split(delim)[:-1] if x != '']
    parts.append(ending)
    return delim.join(parts)


def run_morans(rag, foi):
    """Run morans I, measure of spatial correlation and significance.

    Parameters
    ----------
    rag     :   RAG
                Region adjacency graph. (https://scikit-image.org/docs/stable/auto_examples/segmentation/plot_rag.html)
    foi     :   string
                feature of interest

    Returns
    -------
    Morans I class object.

    """
    get_logger().info("Calculating morans I group statistic for FOI \"%s\"..." % foi)

    # extract weights
    weights = W.from_networkx(rag)

    # extract the feature of interest from the rag
    morans_features = [rag.nodes[nodes_idx][foi] for nodes_idx in list(rag.nodes)]

    morans_i = Moran(morans_features, weights, two_tailed=False)

    get_logger().info("Morans I value: %s " % morans_i.I)
    get_logger().info("Morans I p norm: %s " % morans_i.p_norm)

    return morans_i



import json
import os
from pathlib import Path

import numpy as np
import pandas as pd

from polarityjam.controller.extractor import Extractor
from polarityjam.controller.plotter import Plotter
from polarityjam.controller.segmenter import CellposeSegmenter
from polarityjam.model.collection import PropertiesCollection
from polarityjam.model.parameter import RuntimeParameter, PlotParameter, SegmentationParameter, ImageParameter
from polarityjam.polarityjam_logging import get_logger
from polarityjam.utils.io import read_parameters, read_image, get_tif_list, read_key_file, \
    get_doc_file_prefix, write_dict_to_yml, create_path_recursively


def run(args):
    """Run the polarityjam pipeline.

    Args:
        args:
            The arguments provided by the user.

    """
    # read args
    param_file = args.param
    filepath = args.in_file
    filename = args.filename_prefix
    output_path = args.out_path

    # print info
    get_logger().info(
        "Arguments provided: %s" % json.dumps(
            {"param": param_file, "in_file": filepath, "out_path": output_path, "filename": filename},
            sort_keys=True, indent=4
        )
    )

    # process args
    if not filename:
        filename, _ = os.path.splitext(os.path.basename(filepath))

    parameters = read_parameters(param_file)
    get_logger().info("Parameters: %s" % json.dumps(parameters, sort_keys=True, indent=4))

    # start routine
    _run(filepath, parameters, output_path, filename)
    _finish(parameters, output_path)


def _finish(parameters, output_path):
    # write parameters to disk
    out_param = Path(output_path).joinpath("%s_param%s" % (get_doc_file_prefix(), ".yml"))
    get_logger().info("Writing parameter to disk: %s" % out_param)
    write_dict_to_yml(out_param, parameters)


def _run(infile, param, output_path, fileout_name):
    create_path_recursively(output_path)

    # read input
    img = read_image(infile)
    params_img = ImageParameter(param)

    # inputParams
    params_input = RuntimeParameter(param)

    # plotter
    params_plot = PlotParameter(param)
    p = Plotter(params_plot)

    # segmenter
    params_seg = SegmentationParameter(param)
    s = CellposeSegmenter(params_seg)

    # prepare segmentation and plot
    img_seg, img_seg_params = s.prepare(img, params_img)
    p.plot_channels(img_seg, img_seg_params, output_path, fileout_name, close=True)

    # segment
    mask = s.segment(img_seg, infile)

    # plot cellpose mask
    p.plot_mask(mask, img_seg, img_seg_params, output_path, fileout_name, close=True)

    # feature extraction
    c = PropertiesCollection()
    e = Extractor(params_input)
    e.extract(img, params_img, mask, fileout_name, output_path, c)

    # visualize
    p.plot_collection(c, close=True)

    get_logger().info("Head of created dataset: \n %s" % c.dataset.head())

    # write output
    fileout_base, _ = os.path.splitext(fileout_name)
    fileout_path = Path(output_path).joinpath(fileout_base + ".csv")
    get_logger().info("Writing features to disk: %s" % fileout_path)
    c.dataset.to_csv(str(fileout_path), index=False)

    get_logger().info("File %s done!" % infile)

    return c.dataset, mask


def run_stack(args):
    """Run the polarityjam pipeline on a stack of images.

    Args:
        args:
            The arguments provided by the user.

    """
    # read args
    param_file = args.param
    inpath = args.in_path
    output_path = args.out_path

    # print info
    get_logger().info(
        "Arguments provided: %s" % json.dumps(
            {"param": param_file, "in_path": inpath, "out_path": output_path},
            sort_keys=True, indent=4
        )
    )

    # process
    parameters = read_parameters(param_file)
    get_logger().info("Parameters: %s" % json.dumps(parameters, sort_keys=True, indent=4))

    file_list = get_tif_list(inpath)
    merged_properties_df = pd.DataFrame()

    for filepath in file_list:
        filepath = Path(filepath)
        filename = filepath.stem + filepath.suffix

        if not ((filepath.suffix != ".tif") or (filepath.suffix != ".tiff")):
            continue

        get_logger().info(
            "Processing file with file stem  \"%s\" and file extension: \"%s\"" % (filepath.stem, filepath.suffix)
        )

        # single run
        properties_df, cellpose_mask = _run(filepath, parameters, output_path, filename)

        if merged_properties_df.empty:
            merged_properties_df = properties_df.copy()
        else:
            merged_properties_df = pd.concat([merged_properties_df, properties_df], ignore_index=True)

        merged_file = str(Path(output_path).joinpath("merged_properties.csv"))
        get_logger().info("Writing merged features to disk: %s" % merged_file)
        merged_properties_df.to_csv(merged_file, index=False)

    _finish(parameters, output_path)


def run_key(args):
    """Run the polarityjam pipeline on a stack of images given an additional csv describing the folder structure.

    Args:
        args:
            The arguments provided by the user.

    """
    # read args
    param_file = args.param
    in_path = args.in_path
    inkey = args.in_key
    output_path_base = args.out_path

    # print info
    get_logger().info(
        "Arguments provided: %s" % json.dumps(
            {"param": param_file, "in_key": inkey, "out_path": output_path_base},
            sort_keys=True, indent=4
        )
    )

    # convert
    output_path_base = Path(output_path_base)
    create_path_recursively(output_path_base)
    in_path = Path(in_path)

    # process
    parameters = read_parameters(param_file)
    get_logger().info("Parameters: %s" % json.dumps(parameters, sort_keys=True, indent=4))
    key_file = read_key_file(inkey)

    # empty DF summarizing overall results
    summary_df = pd.DataFrame()
    summary_properties_df = pd.DataFrame()

    offset = 0
    for k, row in key_file.iterrows():
        # current stack input sub folder
        cur_sub_path = str(row['folder_name'])
        if cur_sub_path.startswith(os.path.sep):
            cur_sub_path = cur_sub_path[1:-1]
        input_path = in_path.joinpath(cur_sub_path)

        # current stack output sub-folder
        cur_sub_out_path = str(row["short_name"])
        output_path = output_path_base.joinpath(cur_sub_out_path)

        # empty results dataset for each condition
        merged_properties_df = pd.DataFrame()

        file_list = get_tif_list(input_path)
        for file_index, filepath in enumerate(file_list):
            filepath = Path(filepath)
            filename = filepath.stem + filepath.suffix

            if not ((filepath.suffix != ".tif") or (filepath.suffix != ".tiff")):
                continue

            get_logger().info(
                "Processing file with: file stem  %s and file extension: %s" % (filepath.stem, filepath.suffix)
            )

            # single run
            properties_df, cellpose_mask = _run(filepath, parameters, output_path, filename)

            # append condition
            properties_df["condition"] = row["short_name"]

            if merged_properties_df.empty:
                merged_properties_df = properties_df.copy()
            else:
                merged_properties_df = pd.concat([merged_properties_df, properties_df], ignore_index=True)

            merged_file = str(output_path.joinpath("merged_table_%s" % row["short_name"] + ".csv"))
            get_logger().info("Writing merged features to disk: %s" % merged_file)
            merged_properties_df.to_csv(merged_file, index=False)

            summary_df.at[offset + file_index, "folder_name"] = row["folder_name"]
            summary_df.at[offset + file_index, "short_name"] = row["short_name"]
            summary_df.at[offset + file_index, "filepath"] = filepath
            summary_df.at[offset + file_index, "cell_number"] = len(np.unique(cellpose_mask))


        offset = offset + len(file_list)


        summary_df_path = output_path_base.joinpath("summary_table" + ".csv")
        get_logger().info("Writing summary table to disk: %s" % summary_df_path)
        summary_df.to_csv(str(summary_df_path), index=False)

        if summary_properties_df.empty:
            summary_properties_df = merged_properties_df.copy()
        else:
            summary_properties_df = pd.concat([summary_properties_df, merged_properties_df], ignore_index=True)

        summary_file = str(output_path_base.joinpath("summary_table_properties.csv"))
        get_logger().info("Writing merged features to disk: %s" % summary_file)
        summary_properties_df.to_csv(summary_file, index=False)

        keyfile_path = output_path_base.joinpath("key_file" + ".csv")
        get_logger().info("Writing key file to disk: %s" % keyfile_path)
        key_file.to_csv(str(keyfile_path), index=False)

    _finish(parameters, output_path_base)



import json

import pandas as pd

from polarityjam import RuntimeParameter
from polarityjam.model.image import BioMedicalImage
from polarityjam.model.moran import Moran
from polarityjam.model.properties import SingleCellMarkerProps, SingleCellNucleusProps, SingleCellProps, \
    SingleCellOrganelleProps, SingleCellMarkerNucleiProps, SingleCellMarkerCytosolProps, SingleCellMarkerMembraneProps, \
    SingleCellJunctionProps, NeighborhoodProps


class PropertiesCollection:
    """Collection of single cell properties"""

    def __init__(self):
        self.dataset = pd.DataFrame()
        self.out_path_dict = {}
        self.img_dict = {}
        self.runtime_params_dict = {}
        self._index = 1
        self._reset_index = 1

    def __len__(self):
        return len(self.dataset)

    def current_index(self):
        """Returns the current index of the dataset."""
        return self._index

    def increase_index(self):
        """Increases the index of the dataset."""
        self._index += 1

    def set_reset_index(self):
        """Set the reset index to the current index."""
        self._reset_index = self._index

    def reset_index(self):
        """Resets the index of the dataset."""
        self._index = self._reset_index

    def get_image_by_img_name(self, img_name: str) -> BioMedicalImage:
        """Get the image channel by image name.

        Args:
            img_name:
                image name of which the channel should be returned

        Returns:
            The image channel as numpy array

        """
        return self.img_dict[img_name]

    def get_runtime_params_by_img_name(self, img_name: str) -> RuntimeParameter:
        """Get the feature of interest given the image name.

        Args:
            img_name:
                The image name of which the feature of interest should be returned

        Returns:
            The feature of interest as numpy array

        """
        return self.runtime_params_dict[img_name]

    def get_properties_by_img_name(self, img_name: str, sorted: bool = True) -> pd.DataFrame:
        """Get the properties of the image given its image name.

        Args:
            sorted:
                If true, the properties are sorted by the label of the single cell
            img_name:
                The image name of which the properties should be returned

        Returns:
            The properties as a pandas dataset

        """
        features = self.dataset.loc[self.dataset["filename"] == img_name]
        if sorted:
            features = features.sort_values(by=["label"], ascending=True)
        return features

    def get_out_path_by_name(self, img_name: str) -> str:
        """Get the output path of the image given its image name.

        Args:
            img_name:
                The image name of which the output path should be returned

        Returns:
            The output path as string
        """
        return self.out_path_dict[img_name]

    def add_sc_marker_polarity_props(self, sc_marker_props: SingleCellMarkerProps):
        """Adds specific single cell marker props to the dataset.

        Args:
            sc_marker_props:
                The single cell marker properties

        """
        self.dataset.at[self._index, "marker_mean_expression"] = sc_marker_props.mean_intensity
        self.dataset.at[
            self._index, "marker_sum_expression"] = sc_marker_props.marker_sum_expression
        self.dataset.at[self._index, "marker_centroid_X"] = sc_marker_props.weighted_centroid[1]  # x is second index
        self.dataset.at[self._index, "marker_centroid_Y"] = sc_marker_props.weighted_centroid[0]
        self.dataset.at[
            self._index, "marker_centroid_orientation_rad"] = sc_marker_props.marker_centroid_orientation_rad
        self.dataset.at[
            self._index, "marker_centroid_orientation_deg"] = sc_marker_props.marker_centroid_orientation_deg

        self.dataset.at[
            self._index, "marker_cue_directional_intensity_ratio"] = sc_marker_props.marker_cue_directional_intensity_ratio
        self.dataset.at[
            self._index, "marker_cue_undirectional_intensity_ratio"] = sc_marker_props.marker_cue_undirectional_intensity_ratio

    def add_sc_nucleus_props(self, nucleus_props: SingleCellNucleusProps):
        """Adds specific single cell nucleus properties to the dataset.

        Args:
            nucleus_props:
                The single cell nucleus properties

        """
        self.dataset.at[self._index, "nuc_X"] = nucleus_props.centroid[1]  # x is second index
        self.dataset.at[self._index, "nuc_Y"] = nucleus_props.centroid[0]
        self.dataset.at[
            self._index, "nuc_displacement_orientation_rad"] = nucleus_props.nuc_displacement_orientation_rad
        self.dataset.at[
            self._index, "nuc_displacement_orientation_deg"] = nucleus_props.nuc_displacement_orientation_deg
        self.dataset.at[self._index, "nuc_shape_orientation_rad"] = nucleus_props.nuc_shape_orientation_rad
        self.dataset.at[self._index, "nuc_shape_orientation_deg"] = nucleus_props.nuc_shape_orientation_deg
        self.dataset.at[self._index, "nuc_major_axis_length"] = nucleus_props.major_axis_length
        self.dataset.at[self._index, "nuc_minor_axis_length"] = nucleus_props.minor_axis_length
        self.dataset.at[self._index, "nuc_area"] = nucleus_props.area
        self.dataset.at[self._index, "nuc_perimeter"] = nucleus_props.perimeter
        self.dataset.at[self._index, "nuc_eccentricity"] = nucleus_props.eccentricity
        self.dataset.at[self._index, "nuc_major_to_minor_ratio"] = nucleus_props.nuc_major_to_minor_ratio

    def add_sc_general_props(
            self, filename: str, img_hash: str, connected_component_label: int, sc_props: SingleCellProps):
        """Adds general single cell properties to the dataset, including filename, image hash,
        connected component label and general properties.

        Args:
            filename:
                The filename of the image
            img_hash:
                The hash of the image
            connected_component_label:
                The connected component label of the single cell
            sc_props:
                The single cell properties

        """
        self.dataset.at[self._index, "filename"] = filename
        self.dataset.at[self._index, "img_hash"] = img_hash
        self.dataset.at[self._index, "label"] = connected_component_label
        self.dataset.at[self._index, "cell_X"] = sc_props.centroid[1]  # x is second index
        self.dataset.at[self._index, "cell_Y"] = sc_props.centroid[0]
        self.dataset.at[self._index, "cell_shape_orientation_rad"] = sc_props.cell_shape_orientation_rad
        self.dataset.at[self._index, "cell_shape_orientation_deg"] = sc_props.cell_shape_orientation_deg
        self.dataset.at[self._index, "cell_major_axis_length"] = sc_props.major_axis_length
        self.dataset.at[self._index, "cell_minor_axis_length"] = sc_props.minor_axis_length
        self.dataset.at[self._index, "cell_eccentricity"] = sc_props.eccentricity
        self.dataset.at[self._index, "cell_major_to_minor_ratio"] = sc_props.cell_major_to_minor_ratio
        self.dataset.at[self._index, "cell_area"] = sc_props.area
        self.dataset.at[self._index, "cell_perimeter"] = sc_props.perimeter
        self.dataset.at[self._index, "cell_corner_points"] = json.dumps(sc_props.cell_corner_points.tolist())

    def add_sc_organelle_props(self, organelle_props: SingleCellOrganelleProps):
        """Adds specific single cell organelle properties to the dataset.

        Args:
            organelle_props:
                The single cell organelle properties

        """
        self.dataset.at[self._index, "organelle_X"] = organelle_props.centroid[1]  # x-axis is the second index
        self.dataset.at[self._index, "organelle_Y"] = organelle_props.centroid[0]
        self.dataset.at[self._index, "nuc_organelle_distance"] = organelle_props.nuc_organelle_distance
        self.dataset.at[self._index, "organelle_orientation_rad"] = organelle_props.organelle_orientation_rad
        self.dataset.at[self._index, "organelle_orientation_deg"] = organelle_props.organelle_orientation_deg

    def add_sc_marker_nuclei_props(self, marker_nuc_props: SingleCellMarkerNucleiProps):
        """Adds specific single cell marker-nuclei properties to the dataset.

        Args:
            marker_nuc_props:
                The single cell marker-nuclei properties

        """
        self.dataset.at[self._index, "marker_mean_expression_nuc"] = marker_nuc_props.mean_intensity
        self.dataset.at[self._index, "marker_sum_expression_nuc"] = marker_nuc_props.marker_sum_expression_nuc
        self.dataset.at[self._index, "marker_nucleus_orientation_rad"] = marker_nuc_props.marker_nucleus_orientation_rad
        self.dataset.at[self._index, "marker_nucleus_orientation_deg"] = marker_nuc_props.marker_nucleus_orientation_deg

    def add_sc_marker_nuclei_cytosol_props(self, marker_nuc_cyt_props: SingleCellMarkerCytosolProps):
        """Adds specific single cell marker-nuclei-cytosol properties to the dataset.

        Args:
            marker_nuc_cyt_props:
                The single cell marker-nuclei-cytosol properties

        """
        self.dataset.at[self._index, "marker_mean_expression_cyt"] = marker_nuc_cyt_props.mean_intensity
        self.dataset.at[self._index, "marker_sum_expression_cyt"] = marker_nuc_cyt_props.marker_sum_expression_cyt

    def add_sc_marker_membrane_props(self, marker_membrane_props: SingleCellMarkerMembraneProps):
        """Adds specific single cell marker-membrane properties to the dataset.

        Args:
            marker_membrane_props:
                The single cell marker-membrane properties

        """
        self.dataset.at[self._index, "marker_mean_expression_mem"] = marker_membrane_props.mean_intensity
        self.dataset.at[self._index, "marker_sum_expression_mem"] = marker_membrane_props.marker_sum_expression_mem

    def add_sc_junction_props(self, sc_junction_props: SingleCellJunctionProps):
        """Adds specific single cell junction properties to the dataset.

        Args:
            sc_junction_props:
                The single cell junction properties

        """
        j_centroid_first, j_centroid_second = sc_junction_props.sc_junction_intensity_props.weighted_centroid

        self.dataset.at[self._index, "junction_centroid_X"] = j_centroid_second  # x-axis is the second index
        self.dataset.at[self._index, "junction_centroid_Y"] = j_centroid_first
        self.dataset.at[self._index, "junction_perimeter"] = sc_junction_props.interface_perimeter
        self.dataset.at[self._index, "junction_protein_area"] = sc_junction_props.sc_junction_intensity_props.area
        # dataset.at[index, "junction_fragmented_perimeter"] = sc_junction_props.junction_fragmented_perimeter
        self.dataset.at[
            self._index, "junction_mean_expression"] = sc_junction_props.sc_junction_interface_props.mean_intensity
        self.dataset.at[
            self._index, "junction_protein_intensity"] = sc_junction_props.junction_protein_intensity
        self.dataset.at[
            self._index, "junction_interface_linearity_index"] = sc_junction_props.junction_interface_linearity_index
        self.dataset.at[self._index, "junction_interface_occupancy"] = sc_junction_props.junction_interface_occupancy
        self.dataset.at[
            self._index, "junction_intensity_per_interface_area"] = sc_junction_props.junction_intensity_per_interface_area
        self.dataset.at[self._index, "junction_cluster_density"] = sc_junction_props.junction_cluster_density
        self.dataset.at[
            self._index, "junction_cue_directional_intensity_ratio"] = sc_junction_props.junction_cue_directional_intensity_ratio
        self.dataset.at[
            self._index, "junction_cue_undirectional_intensity_ratio"] = sc_junction_props.junction_cue_undirectional_intensity_ratio

    def add_morans_i_props(self, morans_i: Moran):
        """Adds Moran's I value to the dataset.

        Args:
            morans_i:
                Moran's I object

        """
        self.dataset.at[self._index, "morans_i"] = morans_i.I
        self.dataset.at[self._index, "morans_p_norm"] = morans_i.p_norm

    def add_neighborhood_props(self, neighborhood_props: NeighborhoodProps):
        """Adds neighborhood properties to the dataset.

        Args:
            neighborhood_props:
                The neighborhood properties

        """
        self.dataset.at[self._index, "neighbors_cell"] = neighborhood_props.num_neighbours

        # fill properties for first nearest neighbors
        self.dataset.at[self._index, "neighbors_mean_dif_1st"] = neighborhood_props.mean_dif_first_neighbors
        self.dataset.at[self._index, "neighbors_median_dif_1st"] = neighborhood_props.median_dif_first_neighbors
        self.dataset.at[self._index, "neighbors_stddev_dif_1st"] = neighborhood_props.var_dif_first_neighbors
        self.dataset.at[self._index, "neighbors_range_dif_1st"] = neighborhood_props.range_dif_first_neighbors

        # fill properties for second-nearest neighbors
        self.dataset.at[self._index, "neighbors_mean_dif_2nd"] = neighborhood_props.mean_dif_second_neighbors
        self.dataset.at[self._index, "neighbors_median_dif_2nd"] = neighborhood_props.median_dif_second_neighbors
        self.dataset.at[self._index, "neighbors_stddev_dif_2nd"] = neighborhood_props.var_dif_second_neighbors
        self.dataset.at[self._index, "neighbors_range_dif_2nd"] = neighborhood_props.range_dif_second_neighbors



import time
import unittest
from pathlib import Path

import polarityjam.test.test_config as config
import polarityjam.test.test_feature_extraction as test_feature_extraction
import polarityjam.test.test_integration as test_integration

from polarityjam.polarityjam_logging import get_logger
from polarityjam.utils.io import create_path_recursively


def start_tests(target_folder=None):
    if target_folder:
        config._TARGET_DIR = Path(target_folder)
        create_path_recursively(target_folder)

    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    # unittests
    suite.addTests(loader.loadTestsFromModule(test_feature_extraction))

    # integration tests
    suite.addTests(loader.loadTestsFromModule(test_integration))

    runner = unittest.TextTestRunner(verbosity=3)
    result = runner.run(suite)
    if result.wasSuccessful():
        time.sleep(5)
        get_logger().info("Success")
        exit(0)
    else:
        get_logger().warning("Failed")
        exit(1)


if __name__ == "__main__":
    start_tests()



from __future__ import annotations

from typing import List, Tuple, Callable, Union, Any

import numpy as np
import skimage.filters
from scipy import ndimage as ndi
from skimage.future.graph import RAG

from polarityjam.polarityjam_logging import get_logger


class Mask:
    """Class representing a base mask."""

    def __init__(self, mask: np.ndarray):
        self.data = mask.astype(bool)

    @classmethod
    def from_threshold_otsu(cls, channel: np.ndarray, gaussian_filter=3) -> Mask:
        """Initializes a mask from a channel using Otsu's method."""
        if gaussian_filter is not None:
            img_channel_blur = ndi.gaussian_filter(channel, sigma=3)
        else:
            img_channel_blur = channel
        interior_mask = np.where(img_channel_blur > skimage.filters.threshold_otsu(img_channel_blur), True, False)

        return cls(interior_mask)

    @classmethod
    def empty(cls, shape: Tuple[int, int]) -> Mask:
        """Creates an empty mask of a given shape."""
        return cls(np.zeros(shape))

    def is_count_below_threshold(self, pixel_value: int) -> bool:
        """Checks if the number of pixels in the mask is below a given threshold.

        Args:
            pixel_value:
                The threshold value.

        Returns:
            True if the number of pixels is below the threshold, False otherwise.
        """
        if len(self.data[self.data == 1]) < pixel_value:
            return True
        return False


class BioMedicalMask(Mask):
    """Class representing a single boolean mask."""

    def __init__(self, mask: np.ndarray):
        super().__init__(mask)

    def invert(self) -> BioMedicalMask:
        """Inverts the mask."""
        return BioMedicalMask(np.invert(self.data))

    def operation(self, overlay: BioMedicalMask, operator: Callable) -> BioMedicalMask:
        """Performs an operation on the mask with another mask."""
        return BioMedicalMask(operator(overlay.data.astype(bool), self.data.astype(bool)))

    def get_outline_from_mask(self, width: int = 1) -> BioMedicalMask:
        """Computes outline for a single cell mask.

        Args:
            width:
                The width of the outline.

        Returns:
            The outline mask.

        """

        sc_mask = self.data.astype(bool)
        dilated_mask = ndi.binary_dilation(sc_mask, iterations=width)
        eroded_mask = ndi.binary_erosion(sc_mask, iterations=width)
        outline_mask = np.logical_xor(dilated_mask, eroded_mask)

        return BioMedicalMask(outline_mask)

    def overlay_instance_segmentation(
            self,
            connected_component_mask: BioMedicalInstanceSegmentationMask
    ) -> BioMedicalInstanceSegmentationMask:
        """Overlays an instance segmentation mask on the mask."""
        return BioMedicalInstanceSegmentationMask(self.data * connected_component_mask.data)

    def to_instance_mask(self, instance_label: int = 1) -> BioMedicalInstanceSegmentationMask:
        """Converts the mask to an instance segmentation mask."""
        return BioMedicalInstanceSegmentationMask(self.data * instance_label)

    def mask_background(
            self,
    ) -> BioMedicalMask:
        """Masks the background of the mask."""
        return BioMedicalMask(
            np.ma.masked_where(self.data == False, self.data)
        )

    def combine(self, other_mask: BioMedicalMask):
        return BioMedicalMask(np.logical_and(self.data, other_mask.data))


class BioMedicalInstanceSegmentationMask(Mask):
    """Class representing an instance segmentation mask."""

    def __init__(self, mask: np.ndarray, dtype: Union[str, Any] = int, background_label: Union[int, float] = 0):
        super().__init__(mask)
        self.background_label = background_label
        self.data = mask.astype(dtype)

    def remove_instance(self, instance_label: int) -> BioMedicalInstanceSegmentationMask:
        """Removes a single instance from the mask."""
        return BioMedicalInstanceSegmentationMask(
            np.where(self.data == instance_label, self.background_label, self.data),
            background_label=self.background_label
        )

    def mask_background(
            self, background_label: Union[int, float] = 0, mask_with: Union[
                np.ndarray, BioMedicalInstanceSegmentationMask] = None
    ) -> BioMedicalInstanceSegmentationMask:
        """Masks the background of the mask."""
        if mask_with is None:
            mask_with = self.data
        elif isinstance(mask_with, BioMedicalInstanceSegmentationMask):
            mask_with = mask_with.data

        return BioMedicalInstanceSegmentationMask(
            np.ma.masked_where(self.data == background_label, mask_with), dtype=float,
            background_label=self.background_label
        )

    def get_single_instance_maks(self, instance_label: int) -> BioMedicalMask:
        """Gets the single cell mask given its label.

        Args:
            instance_label:
                The connected component label.

        Returns:
            The single cell mask.
        """
        return BioMedicalMask(
            np.where(self.data == instance_label, True, False)
        )  # convert connected_component_label to True/False mask

    def __len__(self):
        return len(np.unique(self.data))

    def get_labels(self, exclude_background: bool = True) -> np.ndarray:
        """Gets the labels of the mask.

        Args:
            exclude_background:
                Whether to exclude the background label.

        Returns:
            The labels of the mask.

        """
        labels = np.unique(self.data)

        if exclude_background:
            labels = labels[labels != self.background_label]

        return labels

    def to_semantic_mask(self) -> BioMedicalMask:
        """Converts the mask to a semantic (boolean) mask."""
        return BioMedicalMask(self.data.astype(bool))

    def element_mult(
            self, mask: Union[BioMedicalInstanceSegmentationMask, np.ndarray]
    ) -> BioMedicalInstanceSegmentationMask:
        """Performs an element-wise multiplication of the mask with another mask.

        Args:
            mask:
                The mask to multiply with.

        Returns:
            The resulting mask.

        """
        if isinstance(mask, BioMedicalInstanceSegmentationMask):
            mask = mask.data
        return BioMedicalInstanceSegmentationMask(self.data * mask, dtype=float, background_label=self.background_label)

    def element_add(
            self, mask: Union[BioMedicalInstanceSegmentationMask, np.ndarray]
    ) -> BioMedicalInstanceSegmentationMask:
        """Performs an element-wise addition of the mask with another mask.

        Args:
            mask:
                The mask to add with.

        Returns:
            The resulting mask.

        """
        if isinstance(mask, BioMedicalInstanceSegmentationMask):
            mask = mask.data
        return BioMedicalInstanceSegmentationMask(self.data + mask, dtype=float, background_label=self.background_label)

    def scalar_mult(self, scalar: int) -> BioMedicalInstanceSegmentationMask:
        """Performs a scalar multiplication of the mask.

        Args:
            scalar:
                The scalar to multiply with.

        Returns:
            The resulting mask.

        """
        return BioMedicalInstanceSegmentationMask(self.data * scalar, dtype=float,
                                                  background_label=self.background_label)

    def scalar_add(self, scalar: int) -> BioMedicalInstanceSegmentationMask:
        """Performs a scalar addition of the mask.

        Args:
            scalar:
                The scalar to add with.

        Returns:
            The resulting mask.

        """
        return BioMedicalInstanceSegmentationMask(self.data + scalar, dtype=float,
                                                  background_label=self.background_label)

    def relabel(self, new_labels: Union[dict, np.ndarray],
                exclude_background: bool = True) -> BioMedicalInstanceSegmentationMask:
        """Reallocates the labels of the mask.

        Args:
            new_labels:
                A dictionary mapping old labels to new labels.
                Alternatively, a numpy array that assumes consecutive ordered labels.
            exclude_background:
                Whether to exclude the background label from the reallocation.

        Returns:
            New mask object with a reallocated mask.

        """
        l = len(self) - 1 if exclude_background else len(self)

        old_labels = np.unique(self.data)

        if exclude_background:
            old_labels = old_labels[old_labels != self.background_label]

        if isinstance(new_labels, dict):
            new_labels = np.array([new_labels[label] for label in old_labels])

        err_str = f"The number of new labels must match the number of old labels. Got length %s and length %s." % (
            len(new_labels), l
        )
        assert len(new_labels) == l, err_str

        new_mask = np.zeros_like(self.data).astype(new_labels.dtype)
        for old_label, new_label in zip(old_labels, new_labels):
            new_mask[self.data == old_label] = new_label

        return BioMedicalInstanceSegmentationMask(new_mask, dtype=new_mask.dtype,
                                                  background_label=self.background_label)


class BioMedicalInstanceSegmentation:
    """Class representing an entire instance segmentation."""

    def __init__(self, segmentation_mask: BioMedicalInstanceSegmentationMask):
        self.segmentation_mask = segmentation_mask
        self.neighborhood_graph = BioMedicalInstanceSegmentation.get_rag(self.segmentation_mask)
        self.segmentation_mask_connected = self.init_segmentation_mask(
            BioMedicalInstanceSegmentation.get_islands(self.neighborhood_graph))
        self.neighborhood_graph_connected = BioMedicalInstanceSegmentation.get_rag(self.segmentation_mask_connected)

    def update_graphs(self):
        self.neighborhood_graph = BioMedicalInstanceSegmentation.get_rag(self.segmentation_mask)
        self.neighborhood_graph_connected = BioMedicalInstanceSegmentation.get_rag(
            self.segmentation_mask_connected)  # could have islands
        self.segmentation_mask_connected, island_list = self.remove_islands()

        return island_list

    def init_segmentation_mask(self, islands: np.ndarray):
        connected_component_mask = BioMedicalInstanceSegmentationMask(np.copy(self.segmentation_mask.data))

        # remove islands from mask
        for elemet in islands:
            connected_component_mask = connected_component_mask.remove_instance(elemet)

        return connected_component_mask

    def remove_instance_label(self, instance_label):
        """Removes an instance label from the segmentation inplace.

        Args:
            instance_label:
                The instance label to remove.

        Returns:
            All cells that have been additionally removed because they were not connected to the main graph anymore.

        """
        self.neighborhood_graph.remove_node(instance_label)
        self.segmentation_mask = self.segmentation_mask.remove_instance(instance_label)
        self.segmentation_mask_connected = self.segmentation_mask_connected.remove_instance(
            instance_label)

        return self.update_graphs()

    def set_feature_of_interest(self, feature_of_interest_name: str, feature_of_interest_vec: np.ndarray):
        nodes = sorted(self.neighborhood_graph_connected.nodes)

        assert len(feature_of_interest_vec) == len(
            nodes), "The length of the feature of interest vector must match the number of nodes in the graph."

        for idx, node in enumerate(nodes):
            self.neighborhood_graph_connected.nodes[node][feature_of_interest_name] = feature_of_interest_vec[idx]

    def remove_islands(self) -> Tuple[BioMedicalInstanceSegmentationMask, List[int]]:
        """Remove unconnected cells from the mask (Cells without neighbours).

        Returns:
            A tuple containing the connected mask and a list of unconnected cells (Islands).

        """

        # Get list of islands - nodes with no neighbours and remove them
        list_of_islands = BioMedicalInstanceSegmentation.get_islands(self.neighborhood_graph_connected)

        connected_component_mask = BioMedicalInstanceSegmentationMask(np.copy(self.segmentation_mask_connected.data))

        # remove islands from mask
        for elemet in list_of_islands:
            connected_component_mask = connected_component_mask.remove_instance(elemet)

        # remove islands from graph
        for elem in np.unique(list_of_islands):
            self.neighborhood_graph_connected.remove_node(elem)

        get_logger().info("Removed number of islands: %s" % len(list_of_islands))
        get_logger().info("Number of RAG nodes: %s " % len(list(self.neighborhood_graph_connected.nodes)))

        return connected_component_mask, list_of_islands

    @staticmethod
    def get_rag(instance_segmentation: BioMedicalInstanceSegmentationMask, background: int = 0) -> RAG:
        """Gets the RegionAdjacencyGraph for an instance segmentation image.

        Args:
            instance_segmentation:
                The instance segmentation image.
            background:
                The background value.

        Returns:
            The RegionAdjacencyGraph.

        """
        rag = RAG(instance_segmentation.data)
        rag.remove_node(background)
        return rag

    @staticmethod
    def get_islands(neighborhood_graph: RAG) -> np.ndarray:
        list_of_islands = []
        for nodes in neighborhood_graph.nodes:
            if len(list(neighborhood_graph.neighbors(nodes))) == 0:
                list_of_islands.append(nodes)

        return np.unique(list_of_islands)


class SingleCellMasksCollection:
    """Collection of single cell masks and the corresponding label."""

    def __init__(
            self,
            connected_component_label: int,
            sc_mask: BioMedicalMask,
            sc_nucleus_mask: BioMedicalMask,
            sc_organelle_mask: BioMedicalMask,
            sc_membrane_mask: BioMedicalMask,
            sc_cytosol_mask: BioMedicalMask,
            sc_junction_protein_mask: BioMedicalMask,
    ):
        self.connected_component_label = connected_component_label
        self.sc_mask = sc_mask
        self.sc_nucleus_mask = sc_nucleus_mask
        self.sc_organelle_mask = sc_organelle_mask
        self.sc_membrane_mask = sc_membrane_mask
        self.sc_cytosol_mask = sc_cytosol_mask
        self.sc_junction_protein_area_mask = sc_junction_protein_mask



import math
from typing import Tuple, List

import cv2
import numpy as np


def get_contour(sc_img: np.ndarray) -> np.ndarray:
    """Get the contour of a single cell image.

    Args:
        sc_img:
            The single cell image

    Returns:
        The contour coordinates of the single cell

    """
    contours, _ = cv2.findContours(sc_img.astype(bool).astype(np.uint8), cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
    idx_l = 0
    len_l = 0
    if len(contours) > 1:
        # only take the longest contour
        for idx, c in enumerate(contours):
            if len(c) > len_l:
                idx_l = idx
                len_l = len(c)

    contours = contours[idx_l].squeeze(1)

    contours = np.concatenate((contours, [contours[0]]))

    return contours


def get_corner(sc_img: np.ndarray, epsilon: int = 5) -> List[Tuple[int, int]]:
    """Get the corner of a single cell image.

    Args:
        sc_img:
            The single cell image
        epsilon:
            Epsilon value for the Douglas-Peucker algorithm. Determines the maximal perpendicular distance between two
            points to be considered a corner.

    Returns:
        The corner coordinates of the single cell in the image

    """
    contours = get_contour(sc_img)

    corners = douglas_peucker(contours, epsilon)

    return corners[:-1][:]  # last corner point forms circle


def douglas_peucker(points: np.ndarray, epsilon: int) -> np.ndarray:
    """Douglas-Peucker algorithm for determining the contour points of a shape.

    Args:
        points:
            The contour points of the shape
        epsilon:
            Epsilon value for the Douglas-Peucker algorithm. Determines the maximal perpendicular distance between two
            points to be considered a corner.

    Returns:

    """
    if len(points) < 3:
        return points

    dmax = 0
    index = -1
    end = len(points) - 1

    for idx in range(1, end):
        d = perpendicular_distance(points[idx], points[0], points[end])

        if d > dmax:
            index = idx
            dmax = d

    if dmax > epsilon:
        rec_result1 = douglas_peucker(points[0:index + 1], epsilon)
        rec_result2 = douglas_peucker(points[index:], epsilon)

        result = np.concatenate((rec_result1[0:len(rec_result1) - 1], rec_result2), axis=0)

        return result

    else:
        return np.array([points[0], points[end]])


def perpendicular_distance(p, p1, p2):
    """Calculates the perpendicular distance between a point and a line.

    Args:
        p:
            The point to calculate the distance for
        p1:
            The first point of the line
        p2:
            The second point of the line

    Returns:
        The perpendicular distance between the point and the line

    """
    if p1[0] == p2[0]:
        result = abs(p[0] - p1[0])
    else:
        slope = (p2[1] - p1[1]) / (p2[0] - p1[0])
        intercept = p1[1] - (slope * p1[0])

        result = abs(slope * p[0] - p[1] + intercept) / math.sqrt(math.pow(slope, 2) + 1)

    return result



from __future__ import annotations

from typing import Union

import numpy as np
import skimage.measure
from scipy import ndimage as ndi
from skimage.measure._regionprops import RegionProperties

from polarityjam.model.parameter import RuntimeParameter
from polarityjam.model.masks import BioMedicalMask
from polarityjam.compute.compute import compute_reference_target_orientation_rad, \
    compute_angle_deg, compute_marker_vector_norm, compute_shape_orientation_rad, \
    straight_line_length
from polarityjam.compute.corner import get_corner, get_contour
from polarityjam.compute.shape import partition_single_cell_mask
from polarityjam.model.image import BioMedicalChannel


class SingleInstanceProps(RegionProperties):
    """Base class for all single cell properties."""

    def __init__(
            self, single_cell_mask: Union[np.ndarray, BioMedicalMask],
            intensity: Union[np.ndarray, BioMedicalChannel] = None
    ):

        if isinstance(single_cell_mask, BioMedicalMask):
            single_cell_mask = single_cell_mask.to_instance_mask(1).data
        else:
            if not np.issubdtype(single_cell_mask.dtype, np.integer):
                raise RuntimeError("Only integer images allowed!")

        if isinstance(intensity, BioMedicalChannel):
            intensity = intensity.data

        objects = ndi.find_objects(single_cell_mask)

        assert len(objects) == 1, "Only one object allowed!"

        sl = objects[0]

        self.mask = single_cell_mask
        self.intensity = intensity

        super().__init__(sl, 1, single_cell_mask, intensity, True)


class SingleCellProps(SingleInstanceProps):
    """Class representing the properties of a single cell."""

    def __init__(self, single_cell_mask: BioMedicalMask, param: RuntimeParameter):
        self.param = param
        super().__init__(single_cell_mask)

    @property
    def cell_shape_orientation_rad(self):
        # note, the values of orientation from props are in [-pi/2,pi/2] with zero along the y-axis
        return compute_shape_orientation_rad(self.orientation)

    @property
    def cell_shape_orientation_deg(self):
        # note, the values of orientation from props are in [-pi/2,pi/2] with zero along the y-axis
        return compute_angle_deg(self.cell_shape_orientation_rad)

    @property
    def cell_major_to_minor_ratio(self):
        return self.major_axis_length / self.minor_axis_length

    @property
    def cell_corner_points(self):
        return get_corner(self.mask, self.param.dp_epsilon)


class SingleCellNucleusProps(SingleInstanceProps):
    """Class representing the properties of a single nucleus."""

    def __init__(self, single_nucleus_mask: BioMedicalMask, sc_props: SingleCellProps):
        super().__init__(single_nucleus_mask)

        self._sc_props = sc_props

    @property
    def nuc_displacement_orientation_rad(self):
        return compute_reference_target_orientation_rad(
            self._sc_props.centroid[0], self._sc_props.centroid[1], self.centroid[0], self.centroid[1]
        )

    @property
    def nuc_displacement_orientation_deg(self):
        return compute_angle_deg(self.nuc_displacement_orientation_rad)

    @property
    def nuc_shape_orientation_rad(self):
        # note, the values of orientation from props are in [-pi/2,pi/2] with zero along the y-axis
        return compute_shape_orientation_rad(self.orientation)

    @property
    def nuc_shape_orientation_deg(self):
        return compute_angle_deg(self.nuc_shape_orientation_rad)

    @property
    def nuc_major_to_minor_ratio(self):
        return self.major_axis_length / self.minor_axis_length

    @property
    def contour_points(self):
        return get_contour(self.mask)


class SingleCellOrganelleProps(SingleInstanceProps):
    """Class representing the properties of a single organelle."""

    def __init__(self, single_organelle_mask: BioMedicalMask, nucleus_props: SingleCellNucleusProps):
        super().__init__(single_organelle_mask)

        self._nucleus_props = nucleus_props

    @property
    def nuc_organelle_distance(self):
        return compute_marker_vector_norm(
            self.centroid[0], self.centroid[1], self._nucleus_props.centroid[0], self._nucleus_props.centroid[1]
        )

    @property
    def organelle_orientation_rad(self):
        return compute_reference_target_orientation_rad(
            self._nucleus_props.centroid[0], self._nucleus_props.centroid[1], self.centroid[0], self.centroid[1]
        )

    @property
    def organelle_orientation_deg(self):
        return compute_angle_deg(self.organelle_orientation_rad)

    @property
    def contour_points(self):
        return get_contour(self.mask)


class SingleCellMarkerProps(SingleInstanceProps):
    """Class representing the properties of a single cell marker signal."""

    def __init__(self, single_cell_mask: BioMedicalMask, im_marker: BioMedicalChannel, cue_direction: int):
        super().__init__(single_cell_mask, im_marker)
        self.quadrant_masks, self._partition_polygons = partition_single_cell_mask(
            single_cell_mask.data, cue_direction, self.axis_major_length, 4
        )
        self.half_masks, self._partition_polygons = partition_single_cell_mask(
            single_cell_mask.data, cue_direction, self.axis_major_length, 2
        )

    @property
    def marker_centroid_orientation_rad(self):
        return compute_reference_target_orientation_rad(
            self.centroid[0], self.centroid[1], self.weighted_centroid[0], self.weighted_centroid[1]
        )

    @property
    def marker_centroid_orientation_deg(self):
        return compute_angle_deg(self.marker_centroid_orientation_rad)

    @property
    def marker_sum_expression(self):
        return self.mean_intensity * self.area

    @property
    def marker_cue_directional_intensity_ratio(self):
        left = self.intensity * self.half_masks[0] * self.mask
        right = self.intensity * self.half_masks[1] * self.mask

        left_m = np.mean(left)
        right_m = np.mean(right)

        return 1 - 2*left_m/(left_m + right_m)

    @property
    def marker_cue_undirectional_intensity_ratio(self):
        top = self.intensity * self.quadrant_masks[0] * self.mask
        left = self.intensity * self.quadrant_masks[1] * self.mask
        bottom = self.intensity * self.quadrant_masks[2] * self.mask
        right = self.intensity * self.quadrant_masks[3] * self.mask

        return (np.mean(top) + np.mean(bottom)) / (np.mean(left) + np.mean(right) + np.mean(top) + np.mean(bottom))


class SingleCellMarkerMembraneProps(SingleInstanceProps):
    """Class representing the properties of a single cell membrane signal."""

    def __init__(self, single_membrane_mask: BioMedicalMask, im_marker: BioMedicalChannel):
        super().__init__(single_membrane_mask, im_marker)

    @property
    def marker_sum_expression_mem(self):
        return self.mean_intensity * self.area


class SingleCellMarkerNucleiProps(SingleInstanceProps):
    """Class representing the properties of a single cell marker nucleus signal."""

    def __init__(
            self,
            single_nucleus_mask: BioMedicalMask,
            im_marker: BioMedicalChannel,
            sc_nucleus_props: SingleCellNucleusProps,
            sc_marker_props: SingleCellMarkerProps
    ):
        super().__init__(single_nucleus_mask, im_marker)
        self._sc_nucleus_props = sc_nucleus_props
        self._sc_marker_props = sc_marker_props

    @property
    def marker_nucleus_orientation_rad(self):
        return compute_reference_target_orientation_rad(
            self._sc_nucleus_props.centroid[0],
            self._sc_nucleus_props.centroid[1],
            self._sc_marker_props.centroid[0],
            self._sc_marker_props.centroid[1]
        )

    @property
    def marker_nucleus_orientation_deg(self):
        return compute_angle_deg(self.marker_nucleus_orientation_rad)

    @property
    def marker_sum_expression_nuc(self):
        return self.mean_intensity * self.area


class SingleCellMarkerCytosolProps(SingleInstanceProps):
    """Class representing the properties of a single cell marker cytosol signal."""

    def __init__(
            self,
            single_cytosol_mask: BioMedicalMask,
            im_marker: BioMedicalChannel,
            sc_marker_nuclei_props: SingleCellMarkerNucleiProps
    ):
        super().__init__(single_cytosol_mask, im_marker)
        self.sc_marker_nuclei_props = sc_marker_nuclei_props

    @property
    def marker_sum_expression_cyt(self):
        return self.mean_intensity * self.area

    @property
    def marker_cytosol_ratio(self):
        return self.sc_marker_nuclei_props.mean_intensity / self.mean_intensity


class SingleCellJunctionProps:
    # Junction properties subclasses
    class SingleCellJunctionInterfaceProps(SingleInstanceProps):
        # Based on junction mapper: https://doi.org/10.7554/eLife.45413
        def __init__(self, single_membrane_mask: BioMedicalMask, im_junction: BioMedicalChannel):
            super().__init__(single_membrane_mask, im_junction)

    class SingleCellJunctionIntensityProps(SingleInstanceProps):
        def __init__(self, single_junction_intensity_mask: BioMedicalMask, im_junction: BioMedicalChannel):
            super().__init__(single_junction_intensity_mask, im_junction)

    """Class representing the properties of a single cell junction."""

    def __init__(
            self,
            im_junction: BioMedicalChannel,
            single_cell_mask: BioMedicalMask,
            single_cell_membrane_mask: BioMedicalMask,
            single_cell_junction_intensity_mask: BioMedicalMask,
            cue_direction: int,
            dp_epsilon: int,
    ):
        self.im_junction = im_junction
        self.single_cell_mask = single_cell_mask
        self.single_membrane_mask = single_cell_membrane_mask
        self.single_cell_junction_intensity_mask = single_cell_junction_intensity_mask
        self.cue_direction = cue_direction
        self.dp_epsilon = dp_epsilon

        # specific junction properties
        self.sc_junction_interface_props = self.SingleCellJunctionInterfaceProps(single_cell_membrane_mask, im_junction)
        self.sc_junction_intensity_props = self.SingleCellJunctionIntensityProps(
            single_cell_junction_intensity_mask, im_junction
        )

        self.quadrant_masks, self._partition_polygons = partition_single_cell_mask(
            np.logical_or(self.single_cell_mask.data, self.sc_junction_interface_props.mask.data),
            cue_direction,
            self.sc_junction_interface_props.axis_major_length,
            4
        )
        self.half_masks, self._partition_polygons = partition_single_cell_mask(
            np.logical_or(self.single_cell_mask.data, self.sc_junction_interface_props.mask.data),
            cue_direction,
            self.sc_junction_interface_props.axis_major_length,
            2
        )

    @property
    def straight_line_junction_length(self):
        return straight_line_length(get_corner(self.single_cell_mask.data, self.dp_epsilon))

    @property
    def interface_perimeter(self):
        return skimage.measure.perimeter(self.single_cell_mask.data)

    @property
    def junction_interface_linearity_index(self):
        return self.interface_perimeter / self.straight_line_junction_length

    @property
    def junction_interface_occupancy(self):
        return self.sc_junction_intensity_props.area / self.sc_junction_interface_props.area

    @property
    def junction_protein_intensity(self):
        return self.sc_junction_intensity_props.mean_intensity * self.sc_junction_intensity_props.area

    @property
    def junction_intensity_per_interface_area(self):
        return self.junction_protein_intensity / self.sc_junction_interface_props.area

    @property
    def junction_cluster_density(self):
        return self.junction_protein_intensity / self.sc_junction_intensity_props.area

    @property
    def junction_cue_directional_intensity_ratio(self):
        left = self.sc_junction_intensity_props.intensity * self.half_masks[0] * \
               self.sc_junction_intensity_props.mask
        right = self.sc_junction_intensity_props.intensity * self.half_masks[1] * \
                self.sc_junction_intensity_props.mask

        left_m = np.mean(left)
        right_m = np.mean(right)

        return 1 - 2 * left_m / (left_m + right_m)

    @property
    def junction_cue_undirectional_intensity_ratio(self):
        left = self.sc_junction_intensity_props.intensity * \
               self.quadrant_masks[1] * self.sc_junction_intensity_props.mask
        right = self.sc_junction_intensity_props.intensity * \
                self.quadrant_masks[3] * self.sc_junction_intensity_props.mask
        top = self.sc_junction_intensity_props.intensity * \
              self.quadrant_masks[0] * self.sc_junction_intensity_props.mask
        bottom = self.sc_junction_intensity_props.intensity * \
                 self.quadrant_masks[2] * self.sc_junction_intensity_props.mask

        return (np.mean(top) + np.mean(bottom)) / (np.mean(left) + np.mean(right) + np.mean(top) + np.mean(bottom))


class NeighborhoodProps:
    """Class representing the properties of cell neighborhood."""

    def __init__(self):
        self.num_neighbours = None
        self.mean_dif_first_neighbors = 0
        self.median_dif_first_neighbors = 0
        self.var_dif_first_neighbors = 0
        self.range_dif_first_neighbors = 0
        self.mean_dif_second_neighbors = 0
        self.median_dif_second_neighbors = 0
        self.var_dif_second_neighbors = 0
        self.range_dif_second_neighbors = 0


class SingleCellPropertiesCollection:
    """Collection of properties of a single cell."""

    def __init__(
            self, single_cell_props: SingleCellProps,
            nucleus_props: SingleCellNucleusProps,
            organelle_props: SingleCellOrganelleProps,
            marker_props: SingleCellMarkerProps,
            marker_membrane_props: SingleCellMarkerMembraneProps,
            marker_nuc_props: SingleCellMarkerNucleiProps,
            marker_nuc_cyt_props: SingleCellMarkerCytosolProps,
            junction_props: SingleCellJunctionProps
    ):
        self.nucleus_props = nucleus_props
        self.organelle_props = organelle_props
        self.single_cell_props = single_cell_props
        self.marker_nuc_props = marker_nuc_props
        self.marker_nuc_cyt_props = marker_nuc_cyt_props
        self.marker_membrane_props = marker_membrane_props
        self.marker_props = marker_props
        self.junction_props = junction_props



from pathlib import Path
from typing import List, Union

import numpy as np

from polarityjam import RuntimeParameter
from polarityjam.compute.neighborhood import k_neighbor_dif
from polarityjam.model.collection import PropertiesCollection
from polarityjam.model.image import BioMedicalImage, BioMedicalChannel
from polarityjam.model.masks import BioMedicalInstanceSegmentation, \
    SingleCellMasksCollection, BioMedicalInstanceSegmentationMask, BioMedicalMask
from polarityjam.model.moran import Moran, run_morans
from polarityjam.model.properties import SingleCellProps, SingleCellNucleusProps, SingleCellOrganelleProps, \
    SingleCellMarkerProps, SingleCellMarkerMembraneProps, SingleCellMarkerNucleiProps, SingleCellMarkerCytosolProps, \
    SingleCellJunctionProps, SingleCellPropertiesCollection, NeighborhoodProps


class PropertyCollector:
    """Static class, collects features "as they come" in a large dataset. Not responsible for feature calculation!"""

    @staticmethod
    def collect_sc_props(
            sc_prop_collection: SingleCellPropertiesCollection,
            props_collection: PropertiesCollection,
            filename: str,
            img_hash: str,
            connected_component_label: int
    ):

        props_collection.add_sc_general_props(filename, img_hash, connected_component_label,
                                              sc_prop_collection.single_cell_props)

        if sc_prop_collection.marker_props:
            props_collection.add_sc_marker_polarity_props(sc_prop_collection.marker_props)

        if sc_prop_collection.nucleus_props:
            props_collection.add_sc_nucleus_props(sc_prop_collection.nucleus_props)

        if sc_prop_collection.organelle_props:
            props_collection.add_sc_organelle_props(sc_prop_collection.organelle_props)

        if sc_prop_collection.marker_nuc_props:
            props_collection.add_sc_marker_nuclei_props(sc_prop_collection.marker_nuc_props)

        if sc_prop_collection.marker_nuc_cyt_props:
            props_collection.add_sc_marker_nuclei_cytosol_props(sc_prop_collection.marker_nuc_cyt_props)

        if sc_prop_collection.marker_membrane_props:
            props_collection.add_sc_marker_membrane_props(sc_prop_collection.marker_membrane_props)

        if sc_prop_collection.junction_props:
            props_collection.add_sc_junction_props(sc_prop_collection.junction_props)

        props_collection.increase_index()

    @staticmethod
    def collect_group_statistic(props_collection: PropertiesCollection, morans_i: Moran, length: int):
        props_collection.reset_index()
        for i in range(1, length):
            props_collection.add_morans_i_props(morans_i)
            props_collection.increase_index()

    @staticmethod
    def collect_neighborhood_props(props_collection: PropertiesCollection,
                                   neighborhood_props_list: List[NeighborhoodProps]):
        props_collection.reset_index()
        for neighborhood_props in neighborhood_props_list:
            props_collection.add_neighborhood_props(neighborhood_props)
            props_collection.increase_index()

    @staticmethod
    def get_foi(props_collection: PropertiesCollection, foi: str):
        return props_collection.dataset.at[props_collection.current_index() - 1, foi]

    @staticmethod
    def reset_index(props_collection: PropertiesCollection):
        props_collection.reset_index()

    @staticmethod
    def set_reset_index(props_collection: PropertiesCollection):
        props_collection.set_reset_index()

    @staticmethod
    def add_out_path(props_collection: PropertiesCollection, filename: str, path: Union[Path, str]):
        props_collection.out_path_dict[filename] = path

    @staticmethod
    def add_runtime_params(props_collection: PropertiesCollection, filename: str, runtime_params: RuntimeParameter):
        props_collection.runtime_params_dict[filename] = runtime_params

    @staticmethod
    def add_img(props_collection: PropertiesCollection, filename: str, img: BioMedicalImage):
        props_collection.img_dict[filename] = img


class GroupPropertyCollector:
    """Static class, collects group features "as they come" in a large dataset."""

    @staticmethod
    def calc_moran(bio_med_seg: BioMedicalInstanceSegmentation, feature_of_interest_name: str):
        # morans I analysis based on FOI
        morans_i = run_morans(bio_med_seg.neighborhood_graph_connected, feature_of_interest_name)

        return morans_i

    @staticmethod
    def calc_neighborhood(bio_med_seg: BioMedicalInstanceSegmentation, feature_of_interest_name: str):
        return k_neighbor_dif(bio_med_seg.neighborhood_graph_connected, feature_of_interest_name)


class SingleCellPropertyCollector:
    """Static class, collects single cell features "as they come" in a large dataset."""

    @staticmethod
    def calc_sc_props(sc_masks: SingleCellMasksCollection,
                      img: BioMedicalImage, runtime_params: RuntimeParameter) -> SingleCellPropertiesCollection:
        """calculates all properties for the single cell"""

        # properties for single cell
        sc_cell_props = SingleCellPropertyCollector.calc_sc_cell_props(sc_masks.sc_mask, runtime_params)

        # init optional properties
        sc_nuc_props = None
        sc_organelle_props = None
        sc_marker_props = None
        sc_marker_membrane_props = None
        sc_marker_nuclei_props = None
        sc_marker_cytosol_props = None
        sc_junction_props = None

        # properties for nucleus:
        if sc_masks.sc_nucleus_mask is not None:
            sc_nuc_props = SingleCellPropertyCollector.calc_sc_nucleus_props(
                sc_masks.sc_nucleus_mask, sc_cell_props
            )

            # properties for organelle
            if sc_nuc_props and sc_masks.sc_organelle_mask is not None:
                sc_organelle_props = SingleCellPropertyCollector.calc_sc_organelle_props(
                    sc_masks.sc_organelle_mask, sc_nuc_props
                )

        # properties for marker
        if img.marker.data is not None:
            sc_marker_props = SingleCellPropertyCollector.calc_sc_marker_props(
                sc_masks.sc_mask, img.marker.data, runtime_params
            )
            sc_marker_membrane_props = SingleCellPropertyCollector.calc_sc_marker_membrane_props(
                sc_masks.sc_membrane_mask, img.marker.data
            )

            # properties for marker nuclei
            if sc_masks.sc_nucleus_mask is not None:
                sc_marker_nuclei_props = SingleCellPropertyCollector.calc_sc_marker_nuclei_props(
                    sc_masks.sc_nucleus_mask, img.marker.data, sc_nuc_props, sc_marker_props
                )
                sc_marker_cytosol_props = SingleCellPropertyCollector.calc_sc_marker_cytosol_props(
                    sc_masks.sc_cytosol_mask, img.marker, sc_marker_nuclei_props
                )

        if img.junction.data is not None:
            sc_junction_props = SingleCellPropertyCollector.calc_sc_junction_props(
                img.junction,
                sc_masks.sc_mask,
                sc_masks.sc_membrane_mask,
                sc_masks.sc_junction_protein_area_mask,
                runtime_params
            )

        return SingleCellPropertiesCollection(
            sc_cell_props,
            sc_nuc_props,
            sc_organelle_props,
            sc_marker_props,
            sc_marker_membrane_props,
            sc_marker_nuclei_props,
            sc_marker_cytosol_props,
            sc_junction_props
        )

    @staticmethod
    def calc_sc_cell_props(sc_mask: BioMedicalMask, param: RuntimeParameter) -> SingleCellProps:
        return SingleCellProps(sc_mask, param)

    @staticmethod
    def calc_sc_nucleus_props(sc_nucleus_maks: BioMedicalMask, sc_props: SingleCellProps) -> SingleCellNucleusProps:
        return SingleCellNucleusProps(sc_nucleus_maks, sc_props)

    @staticmethod
    def calc_sc_organelle_props(
            sc_organelle_mask: BioMedicalMask,
            sc_nucleus_props: SingleCellNucleusProps
    ) -> SingleCellOrganelleProps:
        return SingleCellOrganelleProps(sc_organelle_mask, sc_nucleus_props)

    @staticmethod
    def calc_sc_marker_props(
            sc_mask: BioMedicalMask, im_marker: BioMedicalChannel, runtime_parameter: RuntimeParameter
    ) -> SingleCellMarkerProps:
        return SingleCellMarkerProps(sc_mask, im_marker, runtime_parameter.cue_direction)

    @staticmethod
    def calc_sc_marker_membrane_props(
            sc_membrane_mask: BioMedicalMask,
            im_marker: BioMedicalChannel
    ) -> SingleCellMarkerMembraneProps:
        return SingleCellMarkerMembraneProps(sc_membrane_mask, im_marker)

    @staticmethod
    def calc_sc_marker_nuclei_props(
            sc_nucleus_mask: BioMedicalMask, im_marker: BioMedicalChannel,
            sc_nucleus_props: SingleCellNucleusProps,
            sc_marker_props: SingleCellMarkerProps
    ) -> SingleCellMarkerNucleiProps:
        return SingleCellMarkerNucleiProps(sc_nucleus_mask, im_marker, sc_nucleus_props, sc_marker_props)

    @staticmethod
    def calc_sc_marker_cytosol_props(
            sc_cytosol_mask: BioMedicalMask, im_marker: BioMedicalChannel,
            sc_marker_nuclei_props: SingleCellMarkerNucleiProps
    ) -> SingleCellMarkerCytosolProps:
        return SingleCellMarkerCytosolProps(sc_cytosol_mask, im_marker, sc_marker_nuclei_props)

    @staticmethod
    def calc_sc_junction_props(
            im_junction: BioMedicalChannel, sc_mask: BioMedicalMask,
            single_cell_membrane_mask: BioMedicalMask,
            single_cell_junction_intensity_mask: BioMedicalMask,
            runtime_parameter: RuntimeParameter
    ) -> SingleCellJunctionProps:

        return SingleCellJunctionProps(
            im_junction,
            single_cell_membrane_mask,
            single_cell_junction_intensity_mask,
            sc_mask,
            runtime_parameter.cue_direction,
            runtime_parameter.dp_epsilon
        )


class SingleCellMaskCollector:

    @staticmethod
    def calc_sc_masks(
            bio_med_img: BioMedicalImage,
            connected_component_label: int,
            membrane_thickness: int,
            nuclei_mask_seg: BioMedicalInstanceSegmentationMask,
            organelle_mask_seg: BioMedicalInstanceSegmentationMask
    ) -> SingleCellMasksCollection:
        sc_mask = bio_med_img.segmentation.segmentation_mask_connected.get_single_instance_maks(
            connected_component_label)
        sc_membrane_mask = sc_mask.get_outline_from_mask(membrane_thickness)

        # init optional sc masks
        sc_nucleus_mask = None
        sc_organelle_mask = None
        sc_cytosol_mask = None
        sc_junction_protein_mask = None

        if nuclei_mask_seg is not None:
            sc_nucleus_mask = nuclei_mask_seg.get_single_instance_maks(connected_component_label)
            sc_cytosol_mask = sc_nucleus_mask.operation(sc_mask, np.logical_xor)

        if organelle_mask_seg is not None:
            sc_organelle_mask = organelle_mask_seg.get_single_instance_maks(connected_component_label)

        if bio_med_img.junction is not None:
            masked_sc_junction_channel = bio_med_img.junction.mask(sc_membrane_mask)
            sc_junction_protein_mask = BioMedicalMask.from_threshold_otsu(
                masked_sc_junction_channel.data,
                gaussian_filter=None
            )

        return SingleCellMasksCollection(
            connected_component_label,
            sc_mask,
            sc_nucleus_mask,
            sc_organelle_mask,
            sc_membrane_mask,
            sc_cytosol_mask,
            sc_junction_protein_mask,
        )



import argparse

from polarityjam.test.test_all import start_tests


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--target-folder',
        required=False,
        help='Specify a folder where the test output is written to. If none is given, temporary folder is taken.',
        default=None
    )
    args = parser.parse_args()

    start_tests(args.target_folder)


if __name__ == "__main__":
    main()
